"""
This module provides string constants for strings used in the Core AaC Spec.

These values aren't expected to change often; it's expected that any deviation
between these values and the core spec will be caught by testing.
"""

# Root Keys
ROOT_KEY_COMMAND_GROUP = "command_group"
ROOT_KEY_ENUM = "enum"
ROOT_KEY_EXTENSION = "ext"
ROOT_KEY_IMPORT = "import"
ROOT_KEY_MODEL = "model"
ROOT_KEY_PLUGIN = "plugin"
ROOT_KEY_SCHEMA = "schema"
ROOT_KEY_SPECIFICATION = "spec"
ROOT_KEY_USECASE = "usecase"

# Common Definition Field Names
DEFINITION_FIELD_ACCEPTANCE = "acceptance"
DEFINITION_FIELD_ACTION = "action"
DEFINITION_FIELD_ADD = "add"
DEFINITION_FIELD_ARGUMENTS = "arguments"
DEFINITION_FIELD_ATTRIBUTES = "attributes"
DEFINITION_FIELD_BEHAVIOR = "behavior"
DEFINITION_FIELD_CHILD = "child"
DEFINITION_FIELD_COMMANDS = "commands"
DEFINITION_FIELD_COMPONENTS = "components"
DEFINITION_FIELD_DEFINITION_SOURCES = "definitionSources"
DEFINITION_FIELD_DEFINITION_VALIDATIONS = "definitionValidations"
DEFINITION_FIELD_DESCRIPTION = "description"
DEFINITION_FIELD_DISPLAY = "display"
DEFINITION_FIELD_EXTENSION_ENUM = "enumExt"
DEFINITION_FIELD_EXTENSION_SCHEMA = "schemaExt"
DEFINITION_FIELD_FIELDS = "fields"
DEFINITION_FIELD_FILES = "files"
DEFINITION_FIELD_GIVEN = "given"
DEFINITION_FIELD_GROUP = "group"
DEFINITION_FIELD_HELP_TEXT = "helpText"
DEFINITION_FIELD_ID = "id"
DEFINITION_FIELD_IDS = "ids"
DEFINITION_FIELD_IMPORT = "import"
DEFINITION_FIELD_INHERITS = "inherits"
DEFINITION_FIELD_INPUT = "input"
DEFINITION_FIELD_NAME = "name"
DEFINITION_FIELD_OUTPUT = "output"
DEFINITION_FIELD_PARENT = "parent"
DEFINITION_FIELD_PARTICIPANTS = "participants"
DEFINITION_FIELD_PRIMITIVE_VALIDATIONS = "primitiveValidations"
DEFINITION_FIELD_REQUIRED = "required"
DEFINITION_FIELD_REQUIREMENTS = "requirements"
DEFINITION_FIELD_ROOT = "root"
DEFINITION_FIELD_SCENARIO = "scenario"
DEFINITION_FIELD_SECTIONS = "sections"
DEFINITION_FIELD_SHALL = "shall"
DEFINITION_FIELD_SOURCE = "source"
DEFINITION_FIELD_STATE = "state"
DEFINITION_FIELD_STEP = "step"
DEFINITION_FIELD_STEPS = "steps"
DEFINITION_FIELD_TAGS = "tags"
DEFINITION_FIELD_TARGET = "target"
DEFINITION_FIELD_THEN = "then"
DEFINITION_FIELD_TYPE = "type"
DEFINITION_FIELD_VALUE = "value"
DEFINITION_FIELD_VALUES = "values"
DEFINITION_FIELD_WHEN = "when"

# Core AaC Definition Names
DEFINITION_NAME_BEHAVIOR = "Behavior"
DEFINITION_NAME_BEHAVIOR_TYPE = "BehaviorType"
DEFINITION_NAME_ENUM = "Enum"
DEFINITION_NAME_EXTENSION = "Extension"
DEFINITION_NAME_FIELD = "Field"
DEFINITION_NAME_MODEL = "Model"
DEFINITION_NAME_PRIMITIVES = "Primitives"
DEFINITION_NAME_REQUIREMENT = "Requirement"
DEFINITION_NAME_REQUIREMENT_REFERENCE = "RequirementReference"
DEFINITION_NAME_SCENARIO = "Scenario"
DEFINITION_NAME_SCHEMA = "Schema"
DEFINITION_NAME_SPECIFICATION = "Specification"
DEFINITION_NAME_USECASE = "Usecase"

# Core AaC Primitive Types
PRIMITIVE_TYPE_BOOL = "bool"
PRIMITIVE_TYPE_DATE = "date"
PRIMITIVE_TYPE_DIRECTORY = "directory"
PRIMITIVE_TYPE_FILE = "file"
PRIMITIVE_TYPE_INT = "int"
PRIMITIVE_TYPE_NUMBER = "number"
PRIMITIVE_TYPE_REFERENCE = "reference"
PRIMITIVE_TYPE_STRING = "string"

# Core AaC Behavior Types
BEHAVIOR_TYPE_PUBLISH_SUBSCRIBE = "PUB_SUB"
BEHAVIOR_TYPE_REQUEST_RESPONSE = "REQUEST_RESPONSE"
BEHAVIOR_TYPE_STARTUP = "STARTUP"
BEHAVIOR_TYPE_TIMER = "TIMER"



"""__init__.py module for the Gen Plugin plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname, abspath
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
)
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.execute.plugin_runner import PluginRunner


from aac.plugins.gen_plugin.gen_plugin_impl import (
    plugin_name,
    gen_plugin,
    before_gen_plugin_check,
    after_gen_plugin_generate,
    gen_project,
    before_gen_project_check,
    after_gen_project_generate,
)


from aac.plugins.check import run_check
from aac.plugins.generate import run_generate

gen_plugin_aac_file_name = "gen_plugin.aac"


def run_gen_plugin(
    aac_plugin_file: str,
    code_output: str,
    test_output: str,
    doc_output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
) -> ExecutionResult:
    """Generate code and stubs for an AaC plugin.  Overwrites will backup existing files."""

    result = ExecutionResult(plugin_name, "gen-plugin", ExecutionStatus.SUCCESS, [])

    gen_plugin_check_result = before_gen_plugin_check(
        aac_plugin_file,
        code_output,
        test_output,
        doc_output,
        no_prompt,
        force_overwrite,
        evaluate,
        run_check,
    )
    if not gen_plugin_check_result.is_success():
        return gen_plugin_check_result
    else:
        result.add_messages(gen_plugin_check_result.messages)

    gen_plugin_result = gen_plugin(
        aac_plugin_file,
        code_output,
        test_output,
        doc_output,
        no_prompt,
        force_overwrite,
        evaluate,
    )
    if not gen_plugin_result.is_success():
        return gen_plugin_result
    else:
        result.add_messages(gen_plugin_result.messages)

    gen_plugin_generate_result = after_gen_plugin_generate(
        aac_plugin_file,
        code_output,
        test_output,
        doc_output,
        no_prompt,
        force_overwrite,
        evaluate,
        run_generate,
    )
    if not gen_plugin_generate_result.is_success():
        return gen_plugin_generate_result
    else:
        result.add_messages(gen_plugin_generate_result.messages)

    return result


def run_gen_project(
    aac_project_file: str,
    output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
) -> ExecutionResult:
    """Generate code and stubs for an AaC project.  Overwrites will backup existing files."""

    result = ExecutionResult(plugin_name, "gen-project", ExecutionStatus.SUCCESS, [])

    gen_project_check_result = before_gen_project_check(
        aac_project_file, output, no_prompt, force_overwrite, evaluate, run_check
    )
    if not gen_project_check_result.is_success():
        return gen_project_check_result
    else:
        result.add_messages(gen_project_check_result.messages)

    gen_project_result = gen_project(
        aac_project_file, output, no_prompt, force_overwrite, evaluate
    )
    if not gen_project_result.is_success():
        return gen_project_result
    else:
        result.add_messages(gen_project_result.messages)

    if not output or len(output) == 0:
        output = dirname(abspath(aac_project_file))
    gen_project_generate_result = after_gen_project_generate(
        aac_project_file, output, no_prompt, force_overwrite, evaluate, run_generate
    )
    if not gen_project_generate_result.is_success():
        return gen_project_generate_result
    else:
        result.add_messages(gen_project_generate_result.messages)

    return result


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    gen_plugin_aac_file = join(dirname(__file__), gen_plugin_aac_file_name)
    definitions = active_context.parse_and_load(gen_plugin_aac_file)

    gen_plugin_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = gen_plugin_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(plugin_definition=gen_plugin_plugin_definition)
    plugin_runner.add_command_callback("gen-plugin", run_gen_plugin)
    plugin_runner.add_command_callback("gen-project", run_gen_project)

    active_context.register_plugin_runner(plugin_runner)



"""The version plugin module."""

from os.path import join, dirname
from aac.plugins.generate.generate_impl import plugin_name, generate, clean
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.execute.plugin_runner import PluginRunner
from aac.execute.aac_execution_result import ExecutionResult, ExecutionStatus

GEN_PLUGIN_AAC_FILE_NAME = "generate.aac"


def run_generate(
    aac_plugin_file: str,
    generator_file: str,
    code_output: str,
    test_output: str,
    doc_output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
) -> ExecutionResult:
    """Run the AaC generate command."""

    result = ExecutionResult(plugin_name, "generate", ExecutionStatus.SUCCESS, [])
    generate_result = generate(
        aac_plugin_file,
        generator_file,
        code_output,
        test_output,
        doc_output,
        no_prompt,
        force_overwrite,
        evaluate,
    )
    if not generate_result.is_success():
        return generate_result
    else:
        result.add_messages(generate_result.messages)

    return result


def run_clean(
    aac_file: str, code_output: str, test_output: str, doc_output: str, no_prompt: bool
) -> ExecutionResult:
    """Run the AaC clean command."""

    result = ExecutionResult(plugin_name, "generate", ExecutionStatus.SUCCESS, [])
    generate_result = clean(aac_file, code_output, test_output, doc_output, no_prompt)
    if not generate_result.is_success():
        return generate_result
    else:
        result.add_messages(generate_result.messages)

    return result


@hookimpl
def register_plugin() -> None:
    """Registers the plugin with the AaC CLI."""

    active_context = LanguageContext()
    generate_aac_file = join(dirname(__file__), GEN_PLUGIN_AAC_FILE_NAME)
    definitions = active_context.parse_and_load(generate_aac_file)

    generate_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = generate_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(plugin_definition=generate_definition)
    plugin_runner.add_command_callback("generate", run_generate)
    plugin_runner.add_command_callback("clean", run_clean)

    active_context.register_plugin_runner(plugin_runner)



"""__init__.py module for the If true then empty plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname
from copy import deepcopy
from typing import Any
from aac.execute.aac_execution_result import (
    ExecutionResult,
)
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from aac.execute.plugin_runner import PluginRunner
from aac.plugins.if_true_then_empty.if_true_then_empty_impl import plugin_name
from aac.plugins.if_true_then_empty.if_true_then_empty_impl import if_true_then_empty


if_true_then_empty_aac_file_name = "if_true_then_empty.aac"


def run_if_true_then_empty(
    instance: Any, definition: Definition, defining_schema, arguments: Any
) -> ExecutionResult:
    """Check to ensure the empty_field_name is empty if bool_field_name is true."""

    argument_copy = deepcopy(arguments)
    args = {}
    for arg in argument_copy:
        args[arg["name"]] = arg["value"]

    return if_true_then_empty(instance, definition, defining_schema, **args)


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    if_true_then_empty_aac_file = join(
        dirname(__file__), if_true_then_empty_aac_file_name
    )
    definitions = active_context.parse_and_load(if_true_then_empty_aac_file)

    if_true_then_empty_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = if_true_then_empty_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(plugin_definition=if_true_then_empty_plugin_definition)

    plugin_runner.add_constraint_callback("If true then empty", run_if_true_then_empty)

    active_context.register_plugin_runner(plugin_runner)



from unittest import TestCase

from aac.context.language_context import LanguageContext
from aac.context.language_error import LanguageError
from aac.context.definition_parser import DefinitionParser
from aac.in_out.parser import parse, ParserError


class TestDefinitionParser(TestCase):
    def test_find_definitions_by_name(self):
        context = LanguageContext()
        definitions = context.get_definitions_by_name("Schema")
        self.assertEqual(len(definitions), 1)
        self.assertEqual(definitions[0].name, "Schema")
        self.assertIsNotNone(definitions[0].instance)

        context.parse_and_load(VALID_AAC_YAML_CONTENT_SPACE_IN_NAME)
        definitions = context.get_definitions_by_name("Test Schema2")
        self.assertEqual(definitions[0].name, "Test Schema2")

    def test_load_definitions_pass(self):
        parser = DefinitionParser()
        context = LanguageContext()
        definitions = parse(VALID_AAC_YAML_CONTENT)
        loaded_definitions = parser.load_definitions(context=context, parsed_definitions=definitions)
        context_definitions = parser.context.get_definitions_by_name(loaded_definitions[0].name)
        name = context_definitions[0].name

        self.assertEqual(name, loaded_definitions[0].name)
        self.assertIsNotNone(context_definitions[0].instance)
        self.assertEqual(len(context_definitions), 1)
        self.assertEqual(loaded_definitions, context_definitions)
        self.assertEqual(len(context_definitions[0].instance.fields), len(loaded_definitions[0].instance.fields))
        self.assertTrue(loaded_definitions[0].source.is_loaded_in_context)

    def test_load_definitions_fail(self):
        parser = DefinitionParser()
        context = LanguageContext()
        with self.assertRaises(ParserError):
            definitions = parse(INVALID_AAC_YAML_CONTENT)
            loaded_definitions = parser.load_definitions(context=context, parsed_definitions=definitions)  # noqa: F841
            self.assertFalse(definitions[0].source.is_loaded_in_context)

    def test_load_incomplete_definition(self):
        parser = DefinitionParser()
        context = LanguageContext()
        with self.assertRaises(ParserError):
            definitions = parse(SCHEMA_WITH_INCOMPLETE_FIELD)
            loaded_definitions = parser.load_definitions(context=context, parsed_definitions=definitions)  # noqa: F841
            self.assertFalse(definitions[0].source.is_loaded_in_context)

    def test_load_bad_enum(self):
        parser = DefinitionParser()
        context = LanguageContext()
        with self.assertRaises(ParserError):
            definitions = parse(SCHEMA_WITH_BAD_ENUM)
            loaded_definitions = parser.load_definitions(context=context, parsed_definitions=definitions)  # noqa: F841
            self.assertFalse(definitions[0].source.is_loaded_in_context)

    def test_load_empty_name(self):
        parser = DefinitionParser()
        context = LanguageContext()
        definition = parse(VALID_AAC_YAML_CONTENT)
        definition[0].name = "1schema_test"
        try:
            loaded_definition = parser.load_definitions(context=context, parsed_definitions=definition)
        except LanguageError as e:
            self.assertNotEqual(e.location, "Unknown location")


VALID_AAC_YAML_CONTENT = """
schema:
  name: TestSchema
  description: |
    This is a test schema.
  fields:
    - name: string_field
      type: string
      description: |
        This is a test field.
    - name: integer_field
      type: integer
      description: |
        This is a test field.
    - name: boolean_field
      type: boolean
      description: |
        This is a test field.
    - name: number_field
      type: number
      description: |
        This is a test field.
""".strip()

VALID_AAC_YAML_CONTENT_SPACE_IN_NAME = """
schema:
  name: Test Schema2
  description: |
    This is a test schema.
  fields:
    - name: string_field
      type: string
      description: |
        This is a test field.
    - name: integer_field
      type: integer
      description: |
        This is a test field.
    - name: boolean_field
      type: boolean
      description: |
        This is a test field.
    - name: number_field
      type: number
      description: |
        This is a test field.
""".strip()

INVALID_AAC_YAML_CONTENT = """
schema:
  description: |
    This is a test schema.
  fields:
    - name: string_field
      type: string
      description: |
        This is a test field.
    - name: integer_field
      type: integer
      description: |
        This is a test field.
    - name: boolean_field
      type: boolean
      description: |
        This is a test field.
    - name: number_field
      type: number
      description: |
        This is a test field.
""".strip()

SCHEMA_WITH_INCOMPLETE_FIELD = """
schema:
  description: |
    This is a test schema.
  fields:
    - name: string_field
      type: string
      description: |
        This is a test field.
    - name: integer_field
      type: integer
      description: |
        This is a test field.
    - name: boolean_field
      type: boolean
      description: |
        This is a test field.
    - name: number_field
      type: number
      description: |
        This is a test field.
    - name: not_used
""".strip()

SCHEMA_WITH_BAD_ENUM= """
schema:
  description: |
    This is a test schema.
  fields:
    - name: string_field
      type: string
      description: |
        This is a test field.
    - name: integer_field
      type: integer
      description: |
        This is a test field.
    - name: boolean_field
      type: boolean
      description: |
        This is a test field.
    - name: number_field
      type: number
      description: |
        This is a test field.
   enum:
    name: bad_enum
    values:
        - bad_value
        - bad_value
""".strip()



"""__init__.py module for the Exclusive fields plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname
from copy import deepcopy
from typing import Any
from aac.execute.aac_execution_result import (
    ExecutionResult,
)
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from aac.execute.plugin_runner import PluginRunner
from aac.plugins.exclusive_fields.exclusive_fields_impl import plugin_name, mutually_exclusive_fields

exclusive_fields_aac_file_name = "exclusive_fields.aac"


def run_mutually_exclusive_fields(
    instance: Any, definition: Definition, defining_schema, arguments: Any
) -> ExecutionResult:
    """Ensure that only one of the fields are defined at any time."""

    argument_copy = deepcopy(arguments)
    args = {}
    for arg in argument_copy:
        args[arg["name"]] = arg["value"]

    return mutually_exclusive_fields(instance, definition, defining_schema, **args)


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    exclusive_fields_aac_file = join(dirname(__file__), exclusive_fields_aac_file_name)
    definitions = active_context.parse_and_load(exclusive_fields_aac_file)

    exclusive_fields_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = exclusive_fields_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(plugin_definition=exclusive_fields_plugin_definition)

    plugin_runner.add_constraint_callback(
        "Mutually exclusive fields", run_mutually_exclusive_fields
    )

    active_context.register_plugin_runner(plugin_runner)



from unittest import TestCase

from aac.context.language_context import LanguageContext
from aac.context.language_error import LanguageError
from aac.in_out.parser._parser_error import ParserError


class TestLanguageContext(TestCase):
    def test_LanguageContext(self):
        context_1 = LanguageContext()
        context_2 = LanguageContext()
        self.assertEqual(context_1, context_2)

    def test_get_aac_core_file_path(self):
        context = LanguageContext()
        self.assertIn("aac.aac", context.get_aac_core_file_path())

    def test_get_aac_core_as_yaml(self):
        context = LanguageContext()
        aac_yaml = context.get_aac_core_as_yaml()
        for definition in context.get_aac_core_definitions():
            self.assertIn(f"name: {definition.name}", aac_yaml)

    def test_get_aac_core_definitions(self):
        context = LanguageContext()
        definitions = context.get_aac_core_definitions()
        self.assertIn("name: Schema", str(definitions))
        self.assertIn("name: Requirement", str(definitions))
        self.assertIn("name: Model", str(definitions))

    def test_is_aac_instance(self):
        context = LanguageContext()
        object = context.create_aac_object("Schema", {})
        self.assertTrue(context.is_aac_instance(object, "Schema"))

    def test_create_aac_object(self):
        context = LanguageContext()
        object = context.create_aac_object("Schema", {})
        self.assertEqual(str(type(object)), "<class 'aac.lang.Schema'>")

    # Something is going wrong with this method, it is failing to create the class. Putting a pin in it.
    # def test_create_aac_enum(self):
    #     context = LanguageContext()
    #     enum = context.create_aac_enum("RequirementVerificationMethod", "TEST")
    #     self.assertEqual(str(type(enum)), "aac.lang.RequirementVerificationMethod")
    #     self.assertEqual(str(enum), "TEST")

    def test_parse_and_load(self):
        context = LanguageContext()
        definitions = context.parse_and_load(VALID_AAC_YAML_CONTENT)
        self.assertEqual(len(definitions), 1)
        self.assertIsNotNone(definitions[0].instance)
        self.assertEqual(definitions[0].name, "TestSchema")
        self.assertEqual(len(definitions[0].instance.fields), 4)

    def test_remove_definitions(self):
        context = LanguageContext()
        loaded_definition = context.parse_and_load(VALID_AAC_YAML_CONTENT)
        name = loaded_definition[0].name
        definition = context.get_definitions_by_name(name)
        self.assertGreater(len(definition), 0)
        context.remove_definitions(definition)
        definition = context.get_definitions_by_name(name)
        self.assertEqual(len(definition), 0)

    def test_get_definitions(self):
        context = LanguageContext()
        definitions = context.get_definitions()
        # check a handful of the core definitions
        definition_names = [definition.name for definition in definitions]
        self.assertIn("Schema", definition_names)
        self.assertIn("Field", definition_names)
        self.assertIn("Primitive", definition_names)
        self.assertIn("Model", definition_names)

    def test_get_definitions_by_name(self):
        context = LanguageContext()
        definitions = context.get_definitions_by_name("Schema")
        self.assertEqual(len(definitions), 1)
        self.assertEqual(definitions[0].name, "Schema")
        self.assertIsNotNone(definitions[0].instance)

        context.parse_and_load(VALID_AAC_YAML_CONTENT_SPACE_IN_NAME)
        definitions = context.get_definitions_by_name("Test Schema2")
        self.assertEqual(definitions[0].name, "Test Schema2")

    def test_get_definitions_by_root(self):
        context = LanguageContext()
        definitions = context.get_definitions_by_root("schema")
        self.assertGreater(len(definitions), 1)
        definition_names = [definition.name for definition in definitions]
        self.assertIn("Schema", definition_names)

    def test_get_defining_schema_for_root(self):
        context = LanguageContext()
        definition = context.get_defining_schema_for_root("schema")
        self.assertEqual(definition.name, "Schema")

    def test_get_plugin_runners(self):
        context = LanguageContext()
        plugin_runners = context.get_plugin_runners()
        self.assertGreater(len(plugin_runners), 0)
        plugin_names = [runner.plugin_definition.name for runner in plugin_runners]
        self.assertIn("Version", plugin_names)

    def test_get_primitives(self):
        context = LanguageContext()
        primitives = context.get_primitives()
        primitive_names = [primitive.name for primitive in primitives]
        # check the core primitives
        self.assertIn("string", primitive_names)
        self.assertIn("int", primitive_names)
        self.assertIn("bool", primitive_names)
        self.assertIn("number", primitive_names)
        self.assertIn("any", primitive_names)
        self.assertIn("date", primitive_names)
        self.assertIn("file", primitive_names)
        self.assertIn("directory", primitive_names)
        self.assertIn("dataref", primitive_names)
        self.assertIn("typeref", primitive_names)

    def test_get_python_type_from_primitive(self):
        context = LanguageContext()
        self.assertEqual(context.get_python_type_from_primitive("string"), "str")
        self.assertEqual(context.get_python_type_from_primitive("int"), "int")
        self.assertEqual(context.get_python_type_from_primitive("bool"), "bool")
        self.assertEqual(context.get_python_type_from_primitive("number"), "float")

    def test_is_extension_of(self):
        context = LanguageContext()
        definition = context.get_definitions_by_name("Schema")
        self.assertTrue(context.is_extension_of(definition[0], "aac.lang", "AacType"))

    def test_get_definitions_of_type(self):
        context = LanguageContext()
        definitions = context.get_definitions_of_type("aac.lang", "AacType")
        self.assertGreater(len(definitions), 1)

    def test_get_values_by_field_chain(self):
        context = LanguageContext()
        test_values = ["schema.fields.name", "enum.values", "primitive.python_type"]
        for val in test_values:
            values = context.get_values_by_field_chain(val)
            self.assertGreater(len(values), 1)

    def test_is_aac_instance_fail(self):
        with self.assertRaises(KeyError):
            context = LanguageContext()
            context.parse_and_load(VALID_AAC_YAML_CONTENT)
            object = context.create_aac_object("TestSchema", {})
            context.remove_definitions(context.get_definitions_by_name("TestSchema"))
            context.is_aac_instance(object, "TestSchema")

    def test_create_aac_object_incorrect_type(self):
        with self.assertRaises(AttributeError):
            context = LanguageContext()
            object = context.create_aac_object("Schema", "Attribute")  # noqa: F841

    # Something is going wrong with this method, it is failing to create the class. Putting a pin in it.
    # def test_create_aac_enum_incorrect_type(self):
    #   with self.assertRaises(AttributeError):
    #       context = LanguageContext()
    #       enum = context.create_aac_enum("RequirementVerificationMethod", 123)

    def test_parse_and_load_non_definition(self):
        with self.assertRaises(ParserError):
            context = LanguageContext()
            context.parse_and_load("definition")

    def test_remove_definitions_fail(self):
        with self.assertRaises(AttributeError):
            context = LanguageContext()
            context.remove_definitions("TestSchema")

    def test_get_python_type_from_primitive_no_type(self):
        context = LanguageContext()
        with self.assertRaises(LanguageError):
            context.get_python_type_from_primitive("type")

    # Should this method return an error of some kind when it finds no definitions of a name?
    def test_get_definitions_by_name_fail(self):
        context = LanguageContext()
        definitions = context.get_definitions_by_name("invalid name")
        self.assertEqual(definitions, [])
        with self.assertRaises(TypeError):
            definitions = context.get_definitions_by_name(123)

    # Should this method return an error of some kind when it finds no definitions of a root?
    def test_get_definitions_by_root_fail(self):
        context = LanguageContext()
        definitions = context.get_definitions_by_root("invalid root")
        self.assertEqual(definitions, [])
        with self.assertRaises(TypeError):
            definitions = context.get_definitions_by_name(123)

    def test_get_defining_schema_for_root_fail(self):
        with self.assertRaises(LanguageError):
            context = LanguageContext()
            definitions = context.get_defining_schema_for_root("Schema")

    def test_get_python_type_from_primitive_fail(self):
        with self.assertRaises(LanguageError):
            context = LanguageContext()
            context.get_python_type_from_primitive("schema")

    def test_is_extension_of_fail(self):
        context = LanguageContext()
        definition = context.get_definitions_by_name("string")
        self.assertFalse(context.is_extension_of(definition[0], "aac.lang", "AacType"))

    def test_get_definitions_of_type_fail(self):
        with self.assertRaises(LanguageError):
            context = LanguageContext()
            context.get_definitions_of_type("invalid_type", "invalid_name")

    def test_get_values_by_field_chain(self):
        context = LanguageContext()
        test_values = ["schema.field.name", "enum.value", "primitives.python_type"]
        for val in test_values:
            values = context.get_values_by_field_chain(val)
            self.assertLess(len(values), 1)


VALID_AAC_YAML_CONTENT = """
schema:
  name: TestSchema
  description: |
    This is a test schema.
  fields:
    - name: string_field
      type: string
      description: |
        This is a test field.
    - name: integer_field
      type: integer
      description: |
        This is a test field.
    - name: boolean_field
      type: boolean
      description: |
        This is a test field.
    - name: number_field
      type: number
      description: |
        This is a test field.
""".strip()

VALID_AAC_YAML_CONTENT_SPACE_IN_NAME = """
schema:
  name: Test Schema2
  description: |
    This is a test schema.
  fields:
    - name: string_field
      type: string
      description: |
        This is a test field.
    - name: integer_field
      type: integer
      description: |
        This is a test field.
    - name: boolean_field
      type: boolean
      description: |
        This is a test field.
    - name: number_field
      type: number
      description: |
        This is a test field.
""".strip()



"""Common utility functions for the context package."""
from aac.context.language_error import LanguageError


def get_python_module_name(package_name: str) -> str:
    """Returns a Python module name, with underscores, from a given package name, which may contain dashes and spaces."""
    if not package_name or package_name == "":
        # we just using default if no package is declared...which really shouldn't happen in core AaC
        return "default"
    result = package_name.replace(" ", "_").replace("-", "_").lower()
    for item in result.split("."):  # expect name to be dot notation
        if not item[0].isalpha():
            raise LanguageError(f"Invalid AaC name: '{item}' within '{package_name}', the first character must be a letter", "Unknown location")
        for sub_item in item.split("_"):  # and name elements may contain _
            if not sub_item.isalnum():  # but things between underscore must be alphanumeric
                raise LanguageError(f"Invalid AaC package name: '{package_name}', the name must be alphanumeric", "Unknown location")
    return result


def get_python_class_name(name: str) -> str:
    """Returns a Python class name from a given name, which may contain dashes and underscores."""
    if not name or name == "":
        raise LanguageError(f"Invalid AaC package name: '{name}', the name must not be empty", "Unknown location")
    if not name[0].isalpha():
        raise LanguageError(f"Invalid AaC name: '{name}', the first character must be a letter", "Unknown location")
    result = name.replace("-", " ").replace("_", " ").replace(" ", "")
    if not result.isalnum():
        raise LanguageError(f"Invalid AaC name: '{name}', the name must be alphanumeric", "Unknown location")
    return result


def get_fully_qualified_name(package_name: str, class_name: str) -> str:
    """Returns the fully qualified name of a class, given the package name and class name."""
    return f"{get_python_module_name(package_name)}.{get_python_class_name(class_name)}"



import os
from parameterized import parameterized
from unittest import TestCase
from aac.in_out.paths import sanitize_filesystem_path

WORKING_TEST_DIR = os.getcwd()

def _get_working_directory_parent_directory(level_of_directories_up: int = 1) -> str:
    """Get the n-level parent directory from the WORKING_TEST_DIR. Useful for testing '../' in paths."""
    path_to_return = WORKING_TEST_DIR
    for i in range(0, level_of_directories_up):
        new_path = os.path.dirname(path_to_return)
        if new_path != path_to_return:
            path_to_return = new_path
        else:
            break

    return path_to_return

class TestPaths(TestCase):
    # Success test block happy path cases
    @parameterized.expand([
        [f"{WORKING_TEST_DIR}/src/aac/test.py",                     f"{WORKING_TEST_DIR}/src/aac/test.py",],
        [f"{WORKING_TEST_DIR}/../src/aac/test.py",                  f"{os.path.dirname(WORKING_TEST_DIR)}/src/aac/test.py"],
        ["src/aac/test.py",                                         f"{WORKING_TEST_DIR}/src/aac/test.py"],
        ["../../src/aac/test.py",                                   f"{_get_working_directory_parent_directory(2)}/src/aac/test.py"],
        [f"{WORKING_TEST_DIR}/../../src/aac/test.py",               f"{_get_working_directory_parent_directory(2)}/src/aac/test.py"],
        [f"{WORKING_TEST_DIR}/src/aac/test.exe%00.py",              f"{WORKING_TEST_DIR}/src/aac/test.exe.py"],
        [f"{WORKING_TEST_DIR}/%2e%2e%2f/src/aac/test.py",           f"{_get_working_directory_parent_directory(1)}/src/aac/test.py"],
        [f"{WORKING_TEST_DIR}/%2e%2e//src/aac/test.py",             f"{_get_working_directory_parent_directory(1)}/src/aac/test.py"],
        [f"{WORKING_TEST_DIR}/..%2f/src/aac/test.py",               f"{_get_working_directory_parent_directory(1)}/src/aac/test.py"],
        [f"{WORKING_TEST_DIR}/%2e%2e%5c/src/aac/test.py",           f"{_get_working_directory_parent_directory(1)}/src/aac/test.py"],
        [f"{WORKING_TEST_DIR}/%2e%2e\/src/aac/test.py",             f"{_get_working_directory_parent_directory(1)}/src/aac/test.py"],
        [f"{WORKING_TEST_DIR}/..%5c\/src/aac/test.py",              f"{_get_working_directory_parent_directory(1)}/src/aac/test.py"],
        [f"{WORKING_TEST_DIR}/%252e%252e%255c\/src/aac/test.py",    f"{_get_working_directory_parent_directory(1)}/src/aac/test.py"],
        [f"{WORKING_TEST_DIR}/..%255c\/src/aac/test.py",            f"{_get_working_directory_parent_directory(1)}/src/aac/test.py"],
    ])
    def test_sequence_success(self, test_path, expected_result):
        self.assertEqual(sanitize_filesystem_path(test_path), expected_result)

    # Expected fail test block for sad path cases
    @parameterized.expand([
        [f"{WORKING_TEST_DIR}/src/aac/AAA.py",                      f"{WORKING_TEST_DIR}/src/aac/BBB.py"],  # Example of input for a fail-test
        [f"{WORKING_TEST_DIR}/src/aac/AAA.py",                      f""],  # Empty directory
        [f"{WORKING_TEST_DIR}/src/aac/AAA.py",                      f"{WORKING_TEST_DIR}/src/aac//AAA.py"],  # Double forward slash test
        [f" ",                                                      f""],  # Space vs empty/null directory
        [f" ",                                                      f" "],  # Tab vs space directory
        [f"{WORKING_TEST_DIR}/src/aac/AAA.py",                      f"{WORKING_TEST_DIR}/src/aac/AAA.py "],  # Trailing space test
        ])
    def test_sequence_fail(self, test_path, expected_result):
        self.assertNotEqual(sanitize_filesystem_path(test_path), expected_result)  # Example of a fail-test assertNotEqual rather than assertEqual




from unittest import TestCase
from typing import Tuple
from click.testing import CliRunner
from aac.execute.command_line import cli, initialize_cli
from aac.execute.aac_execution_result import ExecutionStatus, ExecutionResult
from aac.context.language_context import LanguageContext


from aac.plugins.constraint_assignment_arguments.constraint_assignment_arguments_impl import (
    plugin_name,
)


from aac.plugins.constraint_assignment_arguments.constraint_assignment_arguments_impl import (
    check_arguments_against_constraint_definition,
)


class TestConstraintAssignmentArguments(TestCase):
    def test_check_arguments_against_constraint_definition(self):

        context = LanguageContext()

        schema_constraint = context.get_definitions_by_name("SchemaConstraint")[0]
        if_true_then_empty = schema_constraint.instance.constraints[0]

        if not if_true_then_empty:
            self.fail("Expected if_true_then_empty to be defined but it was not.")

        result: ExecutionResult = check_arguments_against_constraint_definition(if_true_then_empty, schema_constraint, None)

        self.assertTrue(result.is_success(), f"Expected success but failed with message: {result.messages}")


    def test_check_arguments_against_constraint_definition_fail(self):

        context = LanguageContext()
        context.parse_and_load(BAD_DATA)

        schema_constraint = context.get_definitions_by_name("SchemaConstraintBad")[0]
        exclusive_fields = schema_constraint.instance.constraints[0]

        if not exclusive_fields:
            self.fail("Expected exclusive_fields to be defined but it was not.")

        result: ExecutionResult = check_arguments_against_constraint_definition(exclusive_fields, schema_constraint, None)

        self.assertTrue(not result.is_success(), f"Expected failure with message: {result.messages}")
        self.assertEqual(result.messages[0].message, "The Check arguments against constraint definition constraint for Mutually exclusive fields failed because the constraint definition has a required argument named fields that was not found in the constraint assignment.")
        self.assertEqual(result.messages[0].source, "No file to reference")

BAD_DATA = """
schema:
  name: SchemaConstraintBad
  package: aac.lang
  description: |
    The definition of a schema constraint plugin.  Schema constraints perform
    checks against a defined structure within a model based on its schema definition.
    Defining a schema constraint allows for automated structural quality checks
    by running the 'aac check' command against your model.
  fields:
    - name: name
      type: string
      description: |
        The name of the schema constraint rule.
      is_required: true
    - name: description
      type: string
      description: |
        A high level description of the schema constraint rule.
    - name: universal
      type: bool
      description: |
        Indicates that the constraint should be applied to all schemas without explicit assignment.
        This is a convenience so that you don't have to directly assign the constraint to every schema.
        If not included or false, the constraint must be explicitly assigned to a schema.  But be aware
        that universal schema constraints cannot have input arguments.
      default: false
    - name: arguments
      type: Field[]
      description: |
        List of arguments for the constraint.
    - name: acceptance
      type: Feature[]
      description: |
        A list of acceptance test features that describe the expected behavior of the schema constraint.
  constraints:
    - name: Mutually exclusive fields
"""



"""__init__.py module for the AaC primitive constraints plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname
from typing import Any
from aac.execute.aac_execution_result import ExecutionResult
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.in_out.files.aac_file import AaCFile
from aac.execute.plugin_runner import PluginRunner
from aac.context.source_location import SourceLocation


from aac.plugins.aac_primitives.aac_primitive_constraints_impl import plugin_name


from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_bool
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_date
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_directory
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_file
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_string
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_int
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_number
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_dataref
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_typeref


aac_primitive_constraints_aac_file_name = "aac_primitive_constraints.aac"


def run_check_bool(
    value: str,
    type_declaration: str,
    arguments: Any,
    source: AaCFile,
    location: SourceLocation,
) -> ExecutionResult:
    """Verify that a boolen value is True, False, or None.  None is considered False by python, so we allow None as a valid value."""

    return check_bool(value, type_declaration, source, location)


def run_check_date(
    value: str,
    type_declaration: str,
    arguments: Any,
    source: AaCFile,
    location: SourceLocation,
) -> ExecutionResult:
    """Verify that a date value is interpretable as a date."""

    return check_date(value, type_declaration, source, location)


def run_check_directory(
    value: str,
    type_declaration: str,
    arguments: Any,
    source: AaCFile,
    location: SourceLocation,
) -> ExecutionResult:
    """Verify that a directory value is interpretable as a directory."""

    return check_directory(value, type_declaration, source, location)


def run_check_file(
    value: str,
    type_declaration: str,
    arguments: Any,
    source: AaCFile,
    location: SourceLocation,
) -> ExecutionResult:
    """Verify that a file value is interpretable as a file."""

    return check_file(value, type_declaration, source, location)


def run_check_string(
    value: str,
    type_declaration: str,
    arguments: Any,
    source: AaCFile,
    location: SourceLocation,
) -> ExecutionResult:
    """Verify that a string value is interpretable as a string."""

    return check_string(value, type_declaration, source, location)


def run_check_int(
    value: str,
    type_declaration: str,
    arguments: Any,
    source: AaCFile,
    location: SourceLocation,
) -> ExecutionResult:
    """Verify that an integer value is interpretable as an integer."""

    return check_int(value, type_declaration, source, location)


def run_check_number(
    value: str,
    type_declaration: str,
    arguments: Any,
    source: AaCFile,
    location: SourceLocation,
) -> ExecutionResult:
    """Verify that a number value is interpretable as a number."""

    return check_number(value, type_declaration, source, location)


def run_check_dataref(
    value: str,
    type_declaration: str,
    arguments: Any,
    source: AaCFile,
    location: SourceLocation,
) -> ExecutionResult:
    """Verify that a data reference value is interpretable and exists."""

    return check_dataref(value, type_declaration, source, location)


def run_check_typeref(
    value: str,
    type_declaration: str,
    arguments: Any,
    source: AaCFile,
    location: SourceLocation,
) -> ExecutionResult:
    """Verify that a type reference value is interpretable and exists."""

    return check_typeref(value, type_declaration, source, location)


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    aac_primitive_constraints_aac_file = join(
        dirname(__file__), aac_primitive_constraints_aac_file_name
    )
    definitions = active_context.parse_and_load(aac_primitive_constraints_aac_file)

    aac_primitive_constraints_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = aac_primitive_constraints_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(
        plugin_definition=aac_primitive_constraints_plugin_definition
    )

    plugin_runner.add_constraint_callback("Check bool", run_check_bool)
    plugin_runner.add_constraint_callback("Check date", run_check_date)
    plugin_runner.add_constraint_callback("Check directory", run_check_directory)
    plugin_runner.add_constraint_callback("Check file", run_check_file)
    plugin_runner.add_constraint_callback("Check string", run_check_string)
    plugin_runner.add_constraint_callback("Check int", run_check_int)
    plugin_runner.add_constraint_callback("Check number", run_check_number)
    plugin_runner.add_constraint_callback("Check dataref", run_check_dataref)
    plugin_runner.add_constraint_callback("Check typeref", run_check_typeref)

    active_context.register_plugin_runner(plugin_runner)



"""The AaC Print AaC Definitions plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)
from aac.context.language_context import LanguageContext, AAC_LANG_FILE_NAME
from aac.context.definition import Definition
from aac.in_out.parser._parse_source import parse
import yaml
from os.path import join, dirname
from os import linesep
import inspect

plugin_name = "Print AaC Definitions"


def print_defs(core_only: bool) -> ExecutionResult:
    """Print YAML representation of AaC language definitions."""

    messages: list[str] = []
    context: LanguageContext = LanguageContext()
    definitions: list[Definition] = []
    if core_only:
        definitions = parse(
            join(dirname(inspect.getfile(LanguageContext)), AAC_LANG_FILE_NAME)
        )
    else:
        definitions = context.get_definitions()

    for definition in definitions:
        messages.append(yaml.dump(definition.structure))
        messages.append("---")

    output_message = ExecutionMessage(
        linesep.join(messages), MessageLevel.INFO, None, None
    )

    return ExecutionResult(
        plugin_name, "print-defs", ExecutionStatus.SUCCESS, [output_message]
    )



"""Provides command line processing for the aac tool."""

from aac.execute.plugin_runner import AacCommand, AacCommandArgument
from aac.execute.aac_execution_result import ExecutionResult, ExecutionStatus, ExecutionError, OperationCancelled
from pluggy import HookimplMarker, HookspecMarker

PLUGIN_PROJECT_NAME = "aac"
hookimpl = HookimplMarker(PLUGIN_PROJECT_NAME)
hookspec = HookspecMarker(PLUGIN_PROJECT_NAME)

__all__ = (
    "AacCommand",
    "AacCommandArgument",
    hookimpl,
    hookspec,
    "ExecutionResult",
    "ExecutionStatus",
    "ExecutionError",
    "OperationCancelled",
    "get_plugin",
)



"""AaC Module for easy file writing."""
import sys
import logging
from os import path, makedirs

from aac.context.definition import Definition
from aac.in_out.constants import DEFINITION_SEPARATOR
from aac.in_out.paths import sanitize_filesystem_path


def write_file(uri: str, content: str, overwrite: bool = False) -> None:
    """
    Write string content to a file.

    Args:
        uri (str): the file's full uri
        content (str): contents of the file to write
        overwrite (bool): True to overwrite an existing file or false to not.
    """
    sanitized_uri = sanitize_filesystem_path(uri)
    does_file_exist = path.exists(sanitized_uri)
    file_parent_dir = path.dirname(sanitized_uri)

    if not path.exists(file_parent_dir):
        makedirs(file_parent_dir, exist_ok=True)

    if not overwrite and does_file_exist:
        logging.info(f"{sanitized_uri} already exists, skipping write.")
        return

    try:
        with open(sanitized_uri, "w") as file:
            file.writelines(content)
    except IOError as error:
        logging.error(f"Failed to write file {sanitized_uri} due to error: {error}")


def write_definitions_to_file(definitions: list[Definition], file_uri: str, is_user_editable: bool = True) -> None:
    """
    Given a list of definitions, write them to file uri. Updates the source for definitions passed in.

    Args:
        definitions (list[Definition]): The definitions to write to file.
        file_uri (str): The URI of the file to write the definitions to.
        is_user_editable (bool): True if the AaC file can be edited by users.
    """
    def sort_definitions_by_lexeme_line(definition: Definition) -> int:
        line = sys.maxsize

        if definition.lexemes and definition.lexemes[0].source == file_uri:
            line = definition.lexemes[0].location.line

        return line

    definitions.sort(key=sort_definitions_by_lexeme_line)

    file_content = ""
    for definition in definitions:
        definition.source.uri = file_uri
        definition.source.is_user_editable = is_user_editable

        yaml_doc_separator = DEFINITION_SEPARATOR if file_content else ""
        file_content += f"{yaml_doc_separator}{definition.to_yaml()}"

    write_file(file_uri, file_content, True)



"""An exception class representing a language error condition."""
from attr import attrib, attrs, validators


@attrs(slots=True)
class LanguageError(Exception):
    """A base class representing a language error."""
    message: str = attrib(validator=validators.instance_of(str))
    location: str = attrib(validator=validators.instance_of(str))

    def __init__(self, message: str, location: str):
        """
        An init constructor for the Language Error Class.

        Args:
            message (str): The error message.
            location (str): The source location of the error.
        """
        self.message = message
        self.location = location

    def __str__(self) -> str:
        """
        Method for defining how Language Error looks when represented as a string.

        Returns:
            The Language Error represented as a string.
        """
        return f"Error: {self.message}. Location: {self.location}"



from unittest import TestCase
from aac.context.language_context import LanguageContext
from aac.plugins.exclusive_fields.exclusive_fields_impl import mutually_exclusive_fields


class TestExclusiveFields(TestCase):
    def test_mutually_exclusive_fields(self):

        context = LanguageContext()
        one_definition = context.parse_and_load(TEST_EXCLUSIVE_FIELDS)

        # good data tests
        for good_def in [GOOD_DATA_1, GOOD_DATA_2, GOOD_DATA_3, GOOD_DATA_4, GOOD_DATA_5, GOOD_DATA_6]:
            good = context.parse_and_load(good_def)
            result = mutually_exclusive_fields(good[0].instance, one_definition[0], None, ["alpha", "beta", "gamma"])
            self.assertTrue(result.is_success(), msg=f"Expected success for {good[0].name}")
            context.remove_definitions(good)

        # bad data tests
        for bad_def in [BAD_DATA_1, BAD_DATA_2, BAD_DATA_3, BAD_DATA_4, BAD_DATA_5, BAD_DATA_6, BAD_DATA_7, BAD_DATA_8]:
            bad = context.parse_and_load(bad_def)
            result = mutually_exclusive_fields(bad[0].instance, one_definition[0], None, ["alpha", "beta", "gamma"])
            self.assertFalse(result.is_success(), msg=f"Expected failure for {bad[0].name}")
            self.assertEqual(result.messages[0].source, "No file to reference")
            context.remove_definitions(bad)

        context.remove_definitions(one_definition)


TEST_EXCLUSIVE_FIELDS = """
schema:
    name: One
    package: test.exclusive_fields
    root: one
    fields:
        - name: name
          type: string
          is_required: true
        - name: alpha
          type: string
        - name: beta
          type: string
        - name: gamma
          type: string
    constraints:
        - name: Mutually exclusive fields
          arguments:
            fields:
                - alpha
                - beta
                - gamma
"""

GOOD_DATA_1 = """
one:
    name: GoodOne
    alpha: alpha
"""

GOOD_DATA_2 = """
one:
    name: GoodTwo
    beta: beta
"""

GOOD_DATA_3 = """
one:
    name: GoodThree
    gamma: gamma
"""

GOOD_DATA_4 = """
one:
    name: GoodFour
    alpha: alpha
    alpha: alpha  #  Name & field argument duplication is OK
"""

GOOD_DATA_5 = """
one:
    name: GoodFive
    alpha: beta
"""

GOOD_DATA_6 = """
one:
    name: GoodSix
    alpha: beta
    alpha: gamma
"""

BAD_DATA_1 = """
one:
    name: BadOne
    alpha: alpha
    beta: beta
"""

BAD_DATA_2 = """
one:
    name: BadTwo
    alpha: alpha
    gamma: gamma
"""

BAD_DATA_3 = """
one:
    name: BadThree
    beta: beta
    gamma: gamma
"""

BAD_DATA_4 = """
one:
    name: BadFour
    alpha: alpha
    beta: beta
    gamma: gamma
"""

BAD_DATA_5 = """
one:
    name: BadFive
    alpha: alpha
    alpha: #  This intentionally left as blank/whitespace
"""

BAD_DATA_6 = """
one:
    name: BadSix
    beta: beta
    alpha: alpha
"""

BAD_DATA_7 = """
one:
    name: BadSeven
    alpha: #  This intentionally left as blank/whitespace
    alpha: #  This intentionally left as blank/whitespace
"""

BAD_DATA_8 = """
one:
    name: BadEight
    alpha: beta
    beta: gamma
"""



"""Provides classes for representing the results of executing a plugin."""
from enum import Enum, auto, unique
from attr import attrib, attrs, validators, Factory
from os import linesep
from typing import Optional

from aac.context.source_location import SourceLocation


@unique
class ExecutionStatus(Enum):
    """An enumeration that represents status codes for AaC commands to return."""

    SUCCESS = 0
    CONSTRAINT_FAILURE = auto()
    CONSTRAINT_WARNING = auto()
    PARSER_FAILURE = auto()
    PLUGIN_FAILURE = auto()
    OPERATION_CANCELLED = auto()
    GENERAL_FAILURE = auto()


class MessageLevel(Enum):
    """An enumeration that represents the level of a message."""

    DEBUG = auto()
    INFO = auto()
    WARNING = auto()
    ERROR = auto()


@attrs(slots=True, auto_attribs=True)
class ExecutionMessage:
    """Provides a message for the user.

    Attributes:
        message (str): The textual content of the message.
        level (MessageLevel): value of MessageLevel.DEBUG, .INFO, .WARNING, or .ERROR
        source: (str): The file from whence the message came.
        location: (SourceLocation): The col, row info within the source file.
    """

    message: str = attrib(validator=validators.instance_of(str))
    level: MessageLevel = attrib(validator=validators.instance_of(MessageLevel))
    source: Optional[str] = attrib(
        validator=validators.optional(validators.instance_of(str))
    )
    location: Optional[SourceLocation] = attrib(
        validator=validators.optional(validators.instance_of(SourceLocation))
    )


@attrs(slots=True, auto_attribs=True)
class ExecutionResult:
    """Provides information regarding the results of the execution of a plugin.

    Attributes:
        plugin_name (str): The name of the plugin whose results are included.
        plugin_command_name (str): The name of the command which contributed to these results.
        status_code (ExecutionStatus): A status code for the plugin execution.
        messages (list[str]): A list of messages for the user.
    """

    plugin_name: str = attrib(validator=validators.instance_of(str))
    plugin_command_name: str = attrib(validator=validators.instance_of(str))
    status_code: ExecutionStatus = attrib(
        validator=validators.instance_of(ExecutionStatus)
    )
    messages: list[ExecutionMessage] = attrib(
        default=Factory(list), validator=validators.instance_of(list)
    )

    def add_message(self, message: ExecutionMessage) -> None:
        """Add a message to the list of messages."""
        if type(message) is not ExecutionMessage:
            raise TypeError("Message should be of class ExecutionMessage")
        self.messages.append(message)

    def add_messages(self, messages: list[ExecutionMessage]) -> None:
        """Add messages to the list of messages."""
        for message in messages:
            if type(message) is not ExecutionMessage:
                raise TypeError("Message should be of class ExecutionMessage")
        self.messages.extend(messages)

    def is_success(self) -> bool:
        """Return True if the command succeeded; False, otherwise."""
        return self.status_code == ExecutionStatus.SUCCESS

    def get_messages_as_string(self) -> str:
        """Return the output messages as a combined string."""
        result = ""
        for message in self.messages:
            result += message.message + linesep
            if message.source is not None:
                result += f"  Source: {message.source}"
                if message.location is not None:
                    result += f" (Ln {message.location.line}: Col {message.location.column}: Pos {message.location.position}: Spn {message.location.span})"
                result += linesep
        return result


@attrs(slots=True)
class ExecutionError(Exception):
    """A base class representing a plugin error condition.

    Attributes:
        message (str): a textual description of the reason for the exception.
    """

    message: str = attrib(validator=validators.instance_of(str))


@attrs(slots=True)
class OperationCancelled(Exception):
    """A base class representing an cancelled plugin operation condition.

    Attributes:
        message (str): a textual description of the reason for cancellation.
    """

    message: str = attrib(validator=validators.instance_of(str))



"""The AaC Constraint Assignment Arguments plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

# There may be some unused imports depending on the definition of the plugin...but that's ok
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)
from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from typing import Any


plugin_name = "Constraint Assignment Arguments"


def check_arguments_against_constraint_definition(  # noqa: C901
    instance: Any, definition: Definition, defining_schema
) -> ExecutionResult:
    """Business logic for the Check arguments against constraint definition constraint."""

    context = LanguageContext()
    if not context.is_aac_instance(
        instance, "aac.lang.SchemaConstraintAssignment"
    ) and not context.is_aac_instance(
        instance, "aac.lang.PrimitiveConstraintAssignment"
    ):
        # the constraint failed
        error_msg = ExecutionMessage(
            f"The Check arguments against constraint definition constraint for {instance.name} failed.  You may only use this constraint on SchemaConstraintAssignment or PrimitiveConstraintAssignment definitions.  Received {type(instance)}.)",
            MessageLevel.ERROR,
            ("No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri),
            None,
        )
        return ExecutionResult(
            plugin_name,
            "Check arguments against constraint definition",
            ExecutionStatus.GENERAL_FAILURE,
            [error_msg],
        )

    constraint_name = instance.name
    constraint_args: dict = {}
    if instance.arguments is None:
        constraint_args = {}
    elif not isinstance(instance.arguments, list):
        error_msg = ExecutionMessage(
            f"The Check arguments against constraint definition constraint for {instance.name} failed because the assigned arguments were not parsed as a list of entries and cannot be evaluated.  Received {instance.arguments}",
            MessageLevel.ERROR,
            "No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri,
            None,
        )
        return ExecutionResult(
            plugin_name,
            "Check arguments against constraint definition",
            ExecutionStatus.GENERAL_FAILURE,
            [error_msg],
        )
    else:
        for arg in instance.arguments:
            if not isinstance(arg, dict):
                error_msg = ExecutionMessage(
                    f"The Check arguments against constraint definition constraint for {instance.name} failed because the assigned arguments were not parsed as a list of PluginInputValue entries and cannot be evaluated.  Received {instance.arguments}",
                    MessageLevel.ERROR,
                    "No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri,
                    None,
                )
                return ExecutionResult(
                    plugin_name,
                    "Check arguments against constraint definition",
                    ExecutionStatus.GENERAL_FAILURE,
                    [error_msg],
                )
            constraint_args[arg["name"]] = arg["value"]

    constraint_definition = None
    for plugin in context.get_definitions_by_root("plugin"):
        for schema_constraint in plugin.instance.schema_constraints:
            if schema_constraint.name == constraint_name:
                constraint_definition = schema_constraint
                break
        for primitive_constraint in plugin.instance.primitive_constraints:
            if primitive_constraint.name == constraint_name:
                constraint_definition = primitive_constraint
                break

    # Make sure we found the constraint definition
    if constraint_definition is None:
        # the constraint failed
        error_msg = ExecutionMessage(
            f"The Check arguments against constraint definition constraint for {instance.name} failed because the constraint definition could not be found.",
            MessageLevel.ERROR,
            "No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri,
            None,
        )
        return ExecutionResult(
            plugin_name,
            "Check arguments against constraint definition",
            ExecutionStatus.GENERAL_FAILURE,
            [error_msg],
        )

    # Make sure there are no arguments if the constraint definition has no arguments
    if (
        constraint_definition.arguments is None
        or len(constraint_definition.arguments) == 0
    ):
        if constraint_args is not None and constraint_args != {}:
            # the constraint failed
            error_msg = ExecutionMessage(
                f"The Check arguments against constraint definition constraint for {instance.name} failed because the constraint definition does not have arguments, but the constraint assignment does.  Assigned arguments are {constraint_args}.",
                MessageLevel.ERROR,
                "No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri,
                None,
            )
            return ExecutionResult(
                plugin_name,
                "Check arguments against constraint definition",
                ExecutionStatus.GENERAL_FAILURE,
                [error_msg],
            )

    else:  # If there are arguments, then make sure they match the constraint definition
        defined_argument_names = []
        for field in constraint_definition.arguments:
            # make sure required arguments are present
            defined_argument_names.append(field.name)
            if field.is_required and (
                not constraint_args or field.name not in constraint_args
            ):
                # the constraint failed
                error_msg = ExecutionMessage(
                    f"The Check arguments against constraint definition constraint for {instance.name} failed because the constraint definition has a required argument named {field.name} that was not found in the constraint assignment.",
                    MessageLevel.ERROR,
                    "No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri,
                    None,
                )
                return ExecutionResult(
                    plugin_name,
                    "Check arguments against constraint definition",
                    ExecutionStatus.GENERAL_FAILURE,
                    [error_msg],
                )

        # make sure the arguments provided are defined in the constraint definition
        for arg_name in constraint_args:
            if arg_name not in defined_argument_names:
                # the constraint failed
                error_msg = ExecutionMessage(
                    f"The Check arguments against constraint definition constraint for {instance.name} failed because the constraint definition does not have an argument named {arg_name}, but the constraint assignment does.  Argument {arg_name} is not in defined argument names {defined_argument_names}.",
                    MessageLevel.ERROR,
                    "No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri,
                    None,
                )
                return ExecutionResult(
                    plugin_name,
                    "Check arguments against constraint definition",
                    ExecutionStatus.GENERAL_FAILURE,
                    [error_msg],
                )

    return ExecutionResult(
        plugin_name,
        "Check arguments against constraint definition",
        ExecutionStatus.SUCCESS,
        [],
    )



"""Definition class for Architecture-as-Code."""
from attr import Factory, attrib, attrs, validators
from typing import Any
from uuid import UUID, uuid5, NAMESPACE_DNS
import yaml

from aac.in_out.files.aac_file import AaCFile
from aac.context.lexeme import Lexeme
from aac.context.util import get_python_module_name, get_python_class_name, get_fully_qualified_name
from aac.context.language_error import LanguageError


@attrs(hash=False, eq=False)
class Definition:
    """An Architecture-as-Code definition.

    Attributes:
        uid (UUID): A unique identifier for selecting the specific definition.
        name (str): The name of the definition.
        package (str): The package of the definition.
        content (str): The original source textual representation of the definition.
        source (AaCFile): The source document containing the definition.
        lexemes (list[Lexeme]): A list of lexemes for each item in the parsed definition.
        structure (dict): The dictionary representation of the definition.
        instance (Any): A Python class instance of the definition.
    """

    uid: UUID = attrib(init=False, validator=validators.instance_of(UUID))
    name: str = attrib(validator=validators.instance_of(str))
    package: str = attrib(validator=validators.instance_of(str))
    content: str = attrib(validator=validators.instance_of(str))
    source: AaCFile = attrib(validator=validators.instance_of(AaCFile))
    lexemes: list[Lexeme] = attrib(default=Factory(list), validator=validators.instance_of(list))
    structure: dict = attrib(default=Factory(dict), validator=validators.instance_of(dict))
    instance: Any = attrib(default=None)

    def __attrs_post_init__(self):
        """Post-init hook."""
        self.uid = uuid5(NAMESPACE_DNS, str(self.__hash__()))
        if self.is_import():
            self.name = str(self.uid)

    def __hash__(self) -> int:
        """Return the hash of this Definition."""
        return hash(self.get_fully_qualified_name())

    def __eq__(self, obj):
        """Equals function for the Definition."""

        def is_equal() -> bool:
            equal = self.get_fully_qualified_name() == obj.get_fully_qualified_name()
            equal = equal and self.structure == obj.structure
            return equal

        return isinstance(obj, Definition) and is_equal()

    def get_root_key(self) -> str:
        """Get the root key for the definition.

        Returns:
            The root key for the definition.
        """
        return list(self.structure.keys())[0]

    def is_import(self) -> bool:
        """Return True if the definition is an import definition."""
        return self.get_root_key() == "import"

    def get_python_module_name(self) -> str:
        """Return the python module name for the definition."""
        if self.is_import():
            return ""
        return get_python_module_name(self.package)

    def get_python_class_name(self) -> str:
        """Return the python class name for the definition."""
        if self.is_import():
            return ""
        return get_python_class_name(self.name)

    def get_fully_qualified_name(self) -> str:
        """Return the fully qualified name of the definition."""
        if self.is_import():
            return ""
        # this is just the package and name joined with a dot
        try:
            return get_fully_qualified_name(self.package, self.name)
        except LanguageError as e:
            raise LanguageError(e.message, "No file to reference" if not self.source or self.source.uri == "<string>" else self.source.uri)

    def to_yaml(self) -> str:
        """Return a yaml string based on the current state of the definition including extensions."""
        return yaml.dump(self.structure, sort_keys=False)



"""Custom ParserError class that represents errors encountered during the parsing process."""
from attr import Factory, attrib, attrs, validators
from typing import Optional
from yaml.error import YAMLError


@attrs
class ParserError(BaseException):
    """An error that represents a file that could not be parsed."""
    source: str = attrib(validator=validators.instance_of(str))
    errors: list[str] = attrib(default=Factory(list), validator=validators.instance_of(list))
    yaml_error: Optional[YAMLError] = attrib(default=None, validator=validators.optional(validators.instance_of(YAMLError)))



"""Module for shared path functions."""

import logging
import os
import unicodedata
from urllib.parse import unquote


def sanitize_filesystem_path(file_path: str) -> str:
    r"""
    Sanitizes user-provided file path strings to prevent injection and tampering of filesystem paths.

    This function does not determine if a user can write to the path location, it only sanitizes and evaluates the
    path to a full path. The user must still check that the path is a valid destination.

    This function performs the following sanitizing actions:

        * Removes any URL encoding
        * Converts the path to an ASCII encoding
        * Replaces instances of / or \ with os-specific path separators
        * Normalizes the path
        * Determines the realpath if symbolic links are used
        * Ensures that the path is an absolute path

    Path Traversal Security References:

        * OWASP https://owasp.org/www-community/attacks/Path_Traversal

    Args:
        file_path (str): The file path to sanitize

    Returns:
        A sanitized filepath.
    """

    # Loop and unquote until there are no more n-levels of url encoding.
    decoded_url_path = file_path
    while "%" in decoded_url_path:
        unquoted_path = unquote(decoded_url_path)
        if unquoted_path == decoded_url_path:
            break
        else:
            decoded_url_path = unquoted_path

    null_byte_stripped_path = decoded_url_path.replace("\0", "")
    ascii_string_file_path = unicodedata.normalize("NFKD", null_byte_stripped_path).encode("ascii", "replace").decode()

    # Replace instances of \ / with os-specific filepath separator
    os_specific_separator_file_path = ascii_string_file_path.replace("/", os.path.sep)
    os_specific_separator_file_path = os_specific_separator_file_path.replace("\\", os.path.sep)

    normalized_path = os.path.normpath(os_specific_separator_file_path)
    real_path = os.path.realpath(normalized_path)
    abs_norm_path = os.path.abspath(real_path)
    logging.info(f"Sanitized and converted path '{file_path}' to '{abs_norm_path}'.")

    return abs_norm_path


def is_same_file(path1: str, path2: str) -> bool:
    """
    Return whether or not path1 and path2 refer to the same file.

    Args:
        path1 (str): The first path to compare.
        path2 (str): The second path to compare.

    Returns:
        Return True if both paths refer to the same file; False, otherwise.
    """
    return os.path.normcase(path1) == os.path.normcase(path2)



import os
from tempfile import TemporaryDirectory
from unittest import TestCase

from aac.in_out.parser import parse
from aac.in_out.writer import write_definitions_to_file
from aac.context.language_context import LanguageContext

class TestWriter(TestCase):
    
    def test_write_definitions_to_file(self):
        test_file_name = "out.yaml"

        test_model_name = "TestModel"
        test_model_yaml = """
model:
  name: TestModel
  description: This is a test model with no content.
""".strip()
        test_enum_name = "TestEnum"
        test_enum_yaml = """
enum:
  name: TestEnum
  values:
  - v1
""".strip()

        with TemporaryDirectory() as temp_dir:
            test_model = parse(test_model_yaml)[0]
            test_enum = parse(test_enum_yaml)[0]
            test_file_uri = os.path.join(temp_dir, test_file_name)
            write_definitions_to_file([test_model, test_enum], test_file_uri)

            self.assertTrue(os.path.exists(test_file_uri))

            with open(test_file_uri, "r") as file:
                actual_result = file.read()
                self.assertIn(test_model_yaml, actual_result)
                self.assertIn("---", actual_result)
                self.assertIn(test_enum_yaml, actual_result)

    def test_write_definitions_to_file_preserve_order(self):
        test_file_name = "out.yaml"

        context = LanguageContext()
        test_definitions = context.get_aac_core_definitions()

        with TemporaryDirectory() as temp_dir:
            test_file_uri = os.path.join(temp_dir, test_file_name)
            write_definitions_to_file(test_definitions, test_file_uri)

            self.assertTrue(os.path.exists(test_file_uri))

            actual_result_parsed_definitions = parse(test_file_uri)

            self.assertListEqual([d.name for d in test_definitions], [d.name for d in actual_result_parsed_definitions])
            self.assertListEqual([d.source.uri for d in test_definitions], [d.source.uri for d in actual_result_parsed_definitions])
            # Disabling this test for now since the writer doesn't maintain multiline descriptions correctly which causes inconsistent lexeme lines when written to file then read back in.
            # self.assertListEqual([d.lexemes[0].location.line for d in test_definitions], [d.lexemes[0].location.line for d in actual_result_parsed_definitions]) # noqa E800



from unittest import TestCase
from typing import Tuple
from click.testing import CliRunner
from aac.execute.command_line import cli, initialize_cli
from aac.execute.aac_execution_result import ExecutionStatus
from aac.context.language_context import LanguageContext


from aac.plugins.root_schema_must_have_name.root_schema_must_have_name_impl import (
    plugin_name,
)


from aac.plugins.root_schema_must_have_name.root_schema_must_have_name_impl import (
    root_schema_has_name,
)


class TestRootSchemaMustHaveName(TestCase):

    def test_root_schema_has_name(self):
        context = LanguageContext()
        definitions = context.parse_and_load(root_schema_with_name)
        result = root_schema_has_name(definitions[0].instance, definitions[0], context.get_definitions_by_name("Schema")[0].instance)
        self.assertTrue(result.is_success())
        context.remove_definitions(definitions)

    def test_root_schema_has_name_fail(self):
        context = LanguageContext()
        definitions = context.parse_and_load(root_schema_without_name)
        result = root_schema_has_name(definitions[0].instance, definitions[0], context.get_definitions_by_name("Schema")[0].instance)
        self.assertFalse(result.is_success())
        self.assertEqual(result.messages[0].source, "No file to reference")
        context.remove_definitions(definitions)


root_schema_with_name = """
schema:
  name: test_schema
  root: test_root
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
"""

root_schema_without_name = """
schema:
  name: test_schema
  root: test_root
  fields:
    - name: test_field
      type: string
"""



"""__init__.py module for the No Ext for Final plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname
from typing import Any
from aac.execute.aac_execution_result import (
    ExecutionResult,
)
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from aac.execute.plugin_runner import PluginRunner
from aac.plugins.no_ext_for_final.no_ext_for_final_impl import plugin_name
from aac.plugins.no_ext_for_final.no_ext_for_final_impl import no_extension_for_final


no_ext_for_final_aac_file_name = "no_ext_for_final.aac"


def run_no_extension_for_final(
    instance: Any, definition: Definition, defining_schema, arguments: Any
) -> ExecutionResult:
    """Check every schema for extension entries that are marked with a final modifier."""

    return no_extension_for_final(instance, definition, defining_schema)


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    no_ext_for_final_aac_file = join(dirname(__file__), no_ext_for_final_aac_file_name)
    definitions = active_context.parse_and_load(no_ext_for_final_aac_file)

    no_ext_for_final_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = no_ext_for_final_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(plugin_definition=no_ext_for_final_plugin_definition)

    plugin_runner.add_constraint_callback(
        "No Extension for Final", run_no_extension_for_final
    )

    active_context.register_plugin_runner(plugin_runner)



"""The AaC If true then empty plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

# There may be some unused imports depending on the definition of the plugin...but that's ok
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)
from aac.context.definition import Definition
from typing import Any


plugin_name = "If true then empty"


def if_true_then_empty(
    instance: Any,
    definition: Definition,
    defining_schema,
    bool_field_name: str,
    empty_field_name: str,
) -> ExecutionResult:
    """Business logic for the If true then empty constraint."""

    bool_field_value = getattr(instance, bool_field_name, None)
    if not isinstance(bool_field_value, bool):
        # the constraint failed
        error_msg = ExecutionMessage(
            f"The If true then empty constraint for {instance.name} failed because {bool_field_name} is not a boolean.  Received {bool_field_value}.",
            MessageLevel.ERROR,
            "No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri,
            None,
        )
        return ExecutionResult(
            plugin_name,
            "If true then empty",
            ExecutionStatus.GENERAL_FAILURE,
            [error_msg],
        )

    empty_field_value = getattr(instance, empty_field_name, None)

    is_empty_field_empty = (
        empty_field_value is None
        or empty_field_value == ""
        or empty_field_value == []
        or empty_field_value == {}
    )

    if bool_field_value and not is_empty_field_empty:
        # the constraint failed
        error_msg = ExecutionMessage(
            f"The If true then empty constraint for {instance.name} failed because {bool_field_name} is true and {empty_field_name} is not empty.",
            MessageLevel.ERROR,
            "No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri,
            None,
        )
        return ExecutionResult(
            plugin_name,
            "If true then empty",
            ExecutionStatus.GENERAL_FAILURE,
            [error_msg],
        )

    # the constraint passed
    return ExecutionResult(
        plugin_name, "If true then empty", ExecutionStatus.SUCCESS, []
    )



from unittest import TestCase
from aac.context.language_context import LanguageContext
from aac.context.language_error import LanguageError
from aac.in_out.parser._parser_error import ParserError


class TestNoUndefinedFields(TestCase):
    def test_no_extra_fields(self):
        context = LanguageContext()
        with self.assertRaises(LanguageError):
            definitions = context.parse_and_load(BAD_EXTRA_FIELDS)
            context.remove_definitions(definitions)

        with self.assertRaises(LanguageError):
            definitions = context.parse_and_load(BAD_EXTRA_SUB_FIELDS)
            context.remove_definitions(definitions)

    def test_no_undefined_fields(self):
        context = LanguageContext()
        with self.assertRaises(ParserError):
            definitions = context.parse_and_load(EMPTY_FIELDS)



# we've got plenty of good cases in the default language definition, so just test the bad case
BAD_EXTRA_FIELDS = """
schema:
    name: One
    package: test.no_undefined_fields
    root: one
    not_defined_in_field: fail
    fields:
        - name: name
          type: string
          is_required: true
        - name: extra
          type: string
"""

BAD_EXTRA_SUB_FIELDS = """
schema:
    name: One
    package: test.no_undefined_fields
    root: one
    fields:
        - name: name
          type: string
          is_required: true
          not_defined_in_field: fail
        - name: extra
          type: string
"""

EMPTY_FIELDS = """
schema:
    name:
    package:
"""



from os.path import join, dirname
from unittest import TestCase

from aac.context.language_context import LanguageContext
from aac.execute.plugin_runner import AacCommandArgument, AacCommand, PluginRunner


class TestAacCommandArgument(TestCase):
    def test_init(self):
        arg = AacCommandArgument(name="test", description="test description", data_type="str")
        self.assertEqual(arg.name, "test")
        self.assertEqual(arg.description, "test description")
        self.assertEqual(arg.data_type, "str")
        self.assertIsNone(arg.default)

    def test_invalid_name(self):
        with self.assertRaises(TypeError):
            AacCommandArgument(name=123, description="test description", data_type="str")

    def test_invalid_description(self):
        with self.assertRaises(TypeError):
            AacCommandArgument(name="test", description=123, data_type="str")

    def test_invalid_data_type(self):
        with self.assertRaises(TypeError):
            AacCommandArgument(name="test", description="test description", data_type=123)


class TestAacCommand(TestCase):
    def test_init(self):
        arg = AacCommandArgument(name="test", description="test description", data_type="str")
        callback = hello_world
        command = AacCommand(name="test_command", description="test command description", callback=callback, arguments=[arg])

        self.assertEqual(command.name, "test_command")
        self.assertEqual(command.description, "test command description")
        self.assertEqual(command.callback(), "Hello, World!")
        self.assertEqual(len(command.arguments), 1)
        self.assertEqual(command.arguments[0].name, "test")

    def test_invalid_name(self):
        with self.assertRaises(TypeError):
            AacCommand(name=123, description="test description", callback=none_return, arguments=[])

    def test_invalid_description(self):
        with self.assertRaises(TypeError):
            AacCommand(name="test", description=123, callback=none_return, arguments=[])

    def test_invalid_callback(self):
        with self.assertRaises(TypeError):
            AacCommand(name="test", description="test description", callback="not a function", arguments=[])

    def test_invalid_arguments(self):
        with self.assertRaises(TypeError):
            AacCommand(name="test", description="test description", callback=none_return, arguments="not a list")


class TestPluginRunner(TestCase):
    def test_plugin_runner(self):
        aac_file_path = join(dirname(__file__), "my_plugin.aac")
        context = LanguageContext()
        definitions = context.parse_and_load(aac_file_path)

        plugin_runner = PluginRunner(definitions[0], {}, {})
        plugin_runner.add_command_callback("command", command_callback)
        plugin_runner.add_constraint_callback("constraint", constraint_callback)

        command = plugin_runner.get_command_callback("command")
        constraint = plugin_runner.get_constraint_callback("constraint")
        self.assertEqual("Command Callback", command())
        self.assertEqual("Constraint Callback", constraint())
        self.assertEqual("My plugin", plugin_runner.get_plugin_name())

    def test_plugin_runner_list_instead_of_dict(self):
        with self.assertRaises(TypeError):
            aac_file_path = join(dirname(__file__), "my_plugin.aac")
            context = LanguageContext()
            definitions = context.parse_and_load(aac_file_path)

            plugin_runner = PluginRunner(definitions[0], [], [])  # noqa: F841

    def test_plugin_runner_no_callback(self):
        with self.assertRaises(TypeError):
            aac_file_path = join(dirname(__file__), "my_plugin.aac")
            context = LanguageContext()
            definitions = context.parse_and_load(aac_file_path)

            plugin_runner = PluginRunner(definitions[0], {}, {})
            plugin_runner.add_command_callback("command", "Command Callback")

        with self.assertRaises(TypeError):
            aac_file_path = join(dirname(__file__), "my_plugin.aac")
            context = LanguageContext()
            definitions = context.parse_and_load(aac_file_path)

            plugin_runner = PluginRunner(definitions[0], {}, {})
            plugin_runner.add_constraint_callback("constraint", "Constraint Callback")


def command_callback():
    return "Command Callback"


def constraint_callback():
    return "Constraint Callback"


def none_return():
    return None


def hello_world():
    return "Hello, World!"



"""YAML-specific parsing and scanning functions."""

import logging

from yaml import scan, load_all, SafeLoader, Token
from yaml.parser import ParserError as YAMLParserError
from yaml.scanner import ScannerError as YAMLScannerError

from aac.in_out.parser._parser_error import ParserError


def scan_yaml(source: str, content: str) -> list[Token]:
    """
    Parse the YAML string and produce a list of scanning tokens.

    Args:
        source (str): The source of the YAML content. Used to provide better error messages.
        content (str): The previously parsed YAML content to be scanned.

    Returns:
        The scanned YAML parse content.

    Raises:
        If the YAML is invalid, a ParserError is raised.
    """
    try:
        tokens = list(scan(content, Loader=SafeLoader))
    except YAMLScannerError as error:
        error_messages = _yaml_error_messages("scanner", error, content)
        _log_yaml_error(source, error_messages)
        raise ParserError(source, error_messages)
    except Exception as error:
        logging.error(f"Error: {error}. Encountered in: {source}")
        logging.error(f"Content of error: {content}")
        raise ParserError(source, [f"Encountered the following error: {error}"]) from None
    else:
        return tokens


def parse_yaml(source: str, content: str) -> list[dict]:
    """
    Parse content as a YAML string and return the resulting structure.

    Be sure to use the YAML Parser Cache instead of this function.

    Args:
        source (str): The source of the YAML content. Used to provide better error messages.
        content (str): The YAML content to be parsed.

    Returns:
        The parsed YAML content.

    Raises:
        If the YAML is invalid, a ParserError is raised.
        If the model is not a dictionary, a ParserError is raised.
        If the model does not have (at least) a "name" field, a ParserError is raised.
    """
    try:
        models = [model for model in load_all(content, Loader=SafeLoader) if model]
        _error_if_not_yaml(source, content, models)
        _error_if_not_complete(source, content, models)
    except YAMLParserError as error:
        error_messages = _yaml_error_messages("parser", error, content)
        _log_yaml_error(source, error_messages)
        raise ParserError(source, error_messages, error)
    except YAMLScannerError as error:
        error_messages = _yaml_error_messages("scanner", error, content)
        _log_yaml_error(source, error_messages)
        raise ParserError(source, error_messages, error)
    except Exception as error:
        logging.error(f"Error: {error}. Encountered in: {source}")
        logging.error(f"Content of error: {content}")
        raise ParserError(source, [f"Encountered the following error: {error}"]) from None
    else:
        return models


def _error_if_not_yaml(source, content, models):
    """Raise a ParserError if the model is not a YAML model we can parse."""

    def is_model(model):
        """Return True if the model is further parsable."""
        return isinstance(model, dict)

    # Iterate over each model and test if it is considered a valid model.
    if not all(map(is_model, models)):
        raise ParserError(source, ["Provided content was not YAML.  Ensure file exists. ", content])


def _error_if_not_complete(source, content, models):
    """Raise a ParserError if the model is incomplete."""

    def is_import(model):
        """Return True if the model is an import declaration."""
        type, *_ = model.keys()
        return type == "import"

    def assert_definition_has_name(model):
        """Throws a ParserError if the definition doesn't have a name property."""
        type, *_ = model.keys()
        has_name = model.get(type) and model.get(type).get("name")
        if not has_name:
            raise ParserError(source, [f"Definition is missing field 'name': {content}"])

    # Raise an error if any of the loaded YAML models are incomplete.
    models_without_imports = list(filter(lambda m: not is_import(m), models))
    all(map(assert_definition_has_name, models_without_imports))


def _yaml_error_messages(error_type, error, content) -> list[str]:
    error_messages = [
        f"Encountered an invalid YAML with the following {error_type} error: {error.problem}",
        f"Encountered error at line, column: {error.problem_mark.line+1}, {error.problem_mark.column+1}",
        f"Content of the error: {content}",
    ]
    return error_messages


def _log_yaml_error(source, error_messages):
    logging.error(f"YAML error with the following file: {source}")
    for error_message in error_messages:
        logging.error(error_message)



from unittest import TestCase
from aac.context.language_context import LanguageContext

from aac.plugins.unique_root_keys.unique_root_keys_impl import root_key_names_are_unique


class TestUniqueRootKeys(TestCase):
    def test_root_key_names_are_unique(self):
        context = LanguageContext()
        definitions = context.parse_and_load(GOOD_ROOTS)
        result = root_key_names_are_unique(context)
        self.assertTrue(result.is_success())
        context.remove_definitions(definitions)

    def test_root_key_names_are_unique_false(self):
        context = LanguageContext()
        definitions = context.parse_and_load(BAD_ROOTS)
        result = root_key_names_are_unique(context)
        self.assertFalse(result.is_success())
        self.assertIn("is not unique", str(result.messages))
        self.assertEqual(result.messages[0].source, "No file to reference")
        context.remove_definitions(definitions)


GOOD_ROOTS = """
schema:
  name: One
  package: test.root_keys
  root: one
  fields:
    - name: name
      type: string
---
schema:
  name: Two
  package: test.root_keys
  root: two
  fields:
    - name: name
      type: string
"""

BAD_ROOTS = """
schema:
  name: One
  package: test.root_keys
  root: one
  fields:
    - name: name
      type: string
---
schema:
  name: Two
  package: test.root_keys
  root: one
  fields:
    - name: name
      type: string
"""



"""Constants for IO-related items."""
from os import linesep

YAML_DOCUMENT_SEPARATOR = "---"
DEFINITION_SEPARATOR = f"{YAML_DOCUMENT_SEPARATOR}{linesep}"
DEFAULT_SOURCE_URI = "<string>"
YAML_DOCUMENT_EXTENSION = ".yaml"
AAC_DOCUMENT_EXTENSION = ".aac"



"""__init__.py module for the Check AaC plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
)
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.execute.plugin_runner import PluginRunner


from aac.plugins.check.check_aac_impl import plugin_name, check


checkaac_aac_file_name = "check_aac.aac"


def run_check(aac_file, fail_on_warn, verbose) -> ExecutionResult:
    """Perform AaC file quality checks using defined constraints in the AaC models."""

    result = ExecutionResult(plugin_name, "check", ExecutionStatus.SUCCESS, [])

    check_result = check(aac_file, fail_on_warn, verbose)
    if not check_result.is_success():
        return check_result
    else:
        result.add_messages(check_result.messages)

    return result


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    checkaac_aac_file = join(dirname(__file__), checkaac_aac_file_name)
    definitions = active_context.parse_and_load(checkaac_aac_file)

    checkaac_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = checkaac_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(plugin_definition=checkaac_plugin_definition)
    plugin_runner.add_command_callback("check", run_check)

    active_context.register_plugin_runner(plugin_runner)



from unittest import TestCase
from typing import Tuple
from click.testing import CliRunner
from aac.execute.command_line import cli, initialize_cli
from aac.execute.aac_execution_result import ExecutionResult, ExecutionStatus
from aac.context.language_error import LanguageError


from aac.plugins.aac_primitives.aac_primitive_constraints_impl import plugin_name


from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_bool
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_date
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_directory
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_file
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_string
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_int
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_number
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_dataref
from aac.plugins.aac_primitives.aac_primitive_constraints_impl import check_typeref


class TestAaCPrimitiveConstraints(TestCase):

    def test_check_bool(self):
        result = check_bool(True, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, "Check bool should return success for a valid bool value of 'true'")
        result = check_bool(False, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, "Check bool should return success for a valid bool value of 'false'")
        result = check_bool("Not_A_Bool", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, "Check bool should return a constraint failure for a non-bool value of 'Not_A_Bool'")
        self.assertEqual(result.messages[0].source, "No file to reference")
        
    def test_check_date(self):
        # Test valid date
        result = check_date("2022-01-01", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check date should return success for a valid date value of '2022-01-01'")

        # Test invalid date format
        result = check_date("01-01-2022", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check date should return a constraint failure for an invalid date format of '01-01-2022'")
        self.assertEqual(result.messages[0].source, "No file to reference")

        # Test invalid date value
        result = check_date("2022-02-31", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check date should return a constraint failure for an invalid date value of '2022-02-31'")
        self.assertEqual(result.messages[0].source, "No file to reference")

        # Test invalid date type
        result = check_date(123, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check date should return a constraint failure for an invalid date type of '123'")
        self.assertEqual(result.messages[0].source, "No file to reference")

    def test_check_directory(self):
        # Test valid directory
        result = check_directory("/path/to/directory", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check directory should return success for a valid directory path of '/path/to/directory'")

        # Test invalid directory path
        result = check_directory("/path/with+/bad\characters|in/directory", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check directory should return a constraint failure for a nonexistent directory path of '/path/with+/bad\characters|in/directory'")
        self.assertEqual(result.messages[0].source, "No file to reference")

        # Test invalid directory type
        result = check_directory(123, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check directory should return a constraint failure for an invalid directory type of '123'")
        self.assertEqual(result.messages[0].source, "No file to reference")

    def test_check_file(self):
        # Test valid linux directory and file name
        result = check_file("/path/to/directory/my_file.txt", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check file should return success for a valid directory path and file name of '/path/to/directory/my_file.txt'")

        # Test valid windows directory and file name
        result = check_file("c:\\path\\to\\directory\\my_file.txt", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check file should return success for a valid directory path and file name of 'c:\path\\to\directory\my_file.txt'")

        # Test valid relative path and file name
        result = check_file("./path/to/directory/my_file.txt", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check file should return success for a valid directory path and file name of './path/to/directory/my_file.txt'")

        # Test valid relative path and file name
        result = check_file("../path/to/directory/my_file.txt", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check file should return success for a valid directory path and file name of '../path/to/directory/my_file.txt'")

        # Test weird but valid relative path and file name
        result = check_file("../path/to/../directory/my_file.txt", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check file should return success for a valid directory path and file name of '../path/to/../directory/my_file.txt'")

        # Test weird but valid relative path and file name
        result = check_file("my_file.txt", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check file should return success for a valid file only of 'my_file.txt'")

        # Test invalid directory path
        result = check_file("/path/with+/bad\characters|in/directory/my_file.txt", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check file should return a constraint failure for a nonexistent directory path of '/path/with+/bad\characters|in/directory/my_file.txt'")
        self.assertEqual(result.messages[0].source, "No file to reference")

        # Test invalid directory path
        result = check_file("/path/with/bad/characters/in/file/name/m+y|f*ile.txt", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check file should return a constraint failure for a nonexistent directory path of '/path/with/bad/characters/in/file/name/m+y|f*ile.txt'")
        self.assertEqual(result.messages[0].source, "No file to reference")

        # Test invalid directory type
        result = check_directory(123, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check file should return a constraint failure for an invalid directory type of '123'")
        self.assertEqual(result.messages[0].source, "No file to reference")

    def test_check_string(self):
        # Test valid string
        result = check_string("This is a valid string", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check string should return success for a valid string value of 'This is a valid string'")

        # Test invalid string type
        result = check_string(123, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check string should return a constraint failure for an invalid string type of '123'")
        self.assertEqual(result.messages[0].source, "No file to reference")

    def test_check_int(self):
        # Test valid integer
        result = check_int(123, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check int should return success for a valid integer value of '123'")

        # Test valid integer string
        result = check_int("123", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check int should return success for an string integer value of '123'")

        # Test invalid integer value
        result = check_int("abc", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check int should return a constraint failure for an invalid integer value of 'abc'")
        self.assertEqual(result.messages[0].source, "No file to reference")

    def test_check_number(self):
        # Test valid integer
        result = check_number(123, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check number should return success for a valid integer value of '123'")

        # Test valid float
        result = check_number(3.14, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check number should return success for a valid float value of '3.14'")

        # Test valid string integer
        result = check_number("123", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check number should return success for a string integer value of '123'")

        # Test valid string float
        result = check_number("3.14", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check number should return success for a string float value of '3.14'")

        # Test invalid number value
        result = check_number("abc", "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check number should return a constraint failure for an invalid number value of 'abc'")
        self.assertEqual(result.messages[0].source, "No file to reference")

        # Test invalid number type
        result = check_number([1, 2, 3], "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check number should return a constraint failure for an invalid number type of '[1, 2, 3]'")
        self.assertEqual(result.messages[0].source, "No file to reference")

    def test_check_dataref(self):
        # Test valid dataref
        result = check_dataref("description", "dataref(schema.fields.name)", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check dataref should return success for a valid dataref value of 'description' for field 'schema.fields.name'")

        # Test invalid dataref, bad field chain
        result = check_dataref("description", "dataref(schema.not_a_thing.name)", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check dataref should return a constraint failure for an invalid dataref field chain of 'schema.not_a_thing.name'")
        self.assertEqual(result.messages[0].source, "No file to reference")

        # Test invalid dataref, bad target value
        result = check_dataref("not_a_field_name_that_exists", "dataref(schema.fields.name)", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check dataref should return a constraint failure for an invalid dataref target value of 'not_a_field_name_that_exists'")
        self.assertEqual(result.messages[0].source, "No file to reference")

        # Test invalid dataref type
        result = check_dataref(123, "", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check dataref should return a constraint failure for an invalid dataref type of '123'")
        self.assertEqual(result.messages[0].source, "No file to reference")

    def test_check_typeref(self):
         # Test valid typeref
        result = check_typeref("Schema", "typeref(aac.lang.AacType)", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.SUCCESS, f"{plugin_name}: Check typeref should return success for a valid typeref value of 'Schema' for type 'aac.lang.AacType'")

        # Test invalid typeref value
        result = check_typeref("Not_A_Thing", "typeref(aac.lang.AacType)", None, None)
        self.assertEqual(result.status_code, ExecutionStatus.CONSTRAINT_FAILURE, f"{plugin_name}: Check typeref should return a constraint failure for a invalid valid typeref value of 'Not_A_Thing' for type 'aac.lang.AacType'")
        self.assertEqual(result.messages[0].source, "No file to reference")

        # Test invalid typeref target type
        with self.assertRaises(LanguageError, msg=f"{plugin_name}: Check typeref should return a constraint failure for a invalid valid typeref type 'not.a.valid.type'"):
            result = check_typeref("Schema", "typeref(not.a.valid.type)", None, None)




"""Provides a YAML Cache for parsed files or strings."""
import logging
from attr import Factory, attrib, attrs, validators
from hashlib import md5
from os.path import lexists
from typing import Optional
from yaml import Token

from aac.in_out.paths import sanitize_filesystem_path
from aac.in_out.parser._yaml import parse_yaml, scan_yaml

STRING_YAML_SOURCE = "string"


@attrs(hash=False)
class CacheEntry:
    """A cache entry for YAML strings.

    Attributes:
        hash (str): The hash value for the entry.
        yaml_structures (dict[str, Any]): A dictionary of string hashes to yaml dict/maps.
        yaml_tokens (list[Token]): A list of YAML tokens used to scan the content.
        times_accessed (int): The number of times this entry has been accessed.
    """
    hash: str = attrib(validator=validators.instance_of(str))
    yaml_structures: list[dict] = attrib(default=Factory(list), validator=validators.instance_of(list))
    yaml_tokens: list[Token] = attrib(default=Factory(list), validator=validators.instance_of(list))
    times_accessed: int = 0


@attrs
class YamlLFUCache:
    """A Least Frequently Used (LFU) YAML-Parsing cache for strings.

    Attributes:
        capacity (int): The number of cached files/strings before clearing space according to LFU cache behavior.
        cache (dict[str, CacheEntry]): The internal cache data structure -- this is intended to be private, don't access it directly.
    """

    capacity: int = attrib(default=300, validator=validators.instance_of(int))
    # The internal cache is using a dict with sorting O(n log n) rather than a more complex linked list which would be capable of linear time
    cache: dict[str, CacheEntry] = attrib(default=Factory(dict), validator=validators.instance_of(dict))

    def parse_string(self, string: str, source: str = STRING_YAML_SOURCE) -> list[dict]:
        """Parse the YAML string and return the YAML dictionaries."""
        yaml_dicts = self._get_or_parse_string(source, string).yaml_structures
        return yaml_dicts

    def parse_file(self, file_path: str) -> list[dict]:
        """Parse the YAML file and return the YAML dictionaries."""
        yaml_dicts = []
        file_cache_entry = self._get_file_content_cache_entry(file_path)

        if file_cache_entry:
            yaml_dicts = file_cache_entry.yaml_structures

        return yaml_dicts

    def scan_string(self, string: str, source: str = STRING_YAML_SOURCE) -> list[Token]:
        """Parse the YAML string and return the YAML tokens."""
        return self._get_or_parse_string(source, string).yaml_tokens

    def scan_file(self, file_path: str) -> list[Token]:
        """Parse the YAML file and return the YAML tokens."""
        token_list = []
        file_cache_entry = self._get_file_content_cache_entry(file_path)

        if file_cache_entry:
            token_list = file_cache_entry.yaml_tokens

        return token_list

    # Private Methods #

    def _get_file_content_cache_entry(self, file_path: str) -> Optional[CacheEntry]:
        sanitized_file_path = sanitize_filesystem_path(file_path)
        cache_entry = None
        if lexists(sanitized_file_path):
            with open(sanitized_file_path) as yaml_file:
                file_content = yaml_file.read()
                cache_entry = self._get_or_parse_string(sanitized_file_path, file_content)
        else:
            logging.error(f"Can't parse the file '{sanitized_file_path}' because it doesn't exist.")

        return cache_entry

    def _get_entries_sorted_by_access_count(self) -> list[CacheEntry]:
        return sorted(list(self.cache.values()), key=lambda x: x.times_accessed, reverse=True)

    def _get_entry(self, key: str) -> Optional[CacheEntry]:
        entry_to_return = self.cache.get(key)
        if entry_to_return:
            entry_to_return.times_accessed += 1

        return entry_to_return

    def _put_entry(self, key: str, value: CacheEntry) -> None:

        if len(self.cache) >= self.capacity:
            entry_to_pop, *_ = self._get_entries_sorted_by_access_count()
            self.cache.pop(entry_to_pop.hash)
            logging.debug(f"The YAML parser cache has hit its limits, clearing space for a new entry. Current capacity: {self.capacity}")

        self.cache[key] = value

    def _get_or_parse_string(self, content_source: str, content_string: str) -> CacheEntry:
        """Get the cached entry or parse the string and cache it."""
        content_hash = str(md5(content_string.encode()).hexdigest())
        cache_entry = self._get_entry(content_hash)

        if not cache_entry:
            yaml_dicts = parse_yaml(content_source, content_string)
            yaml_tokens = scan_yaml(content_source, content_string)
            cache_entry = CacheEntry(content_hash, yaml_dicts, yaml_tokens)
            self._put_entry(cache_entry.hash, cache_entry)

        return cache_entry



from unittest import TestCase
from aac.context.language_context import LanguageContext

from aac.plugins.if_true_then_empty.if_true_then_empty_impl import if_true_then_empty


class TestIfTrueThenEmpty(TestCase):
    def test_if_true_then_empty(self):

        context = LanguageContext()

        test_schema = context.parse_and_load(TEST_SCHEMA)

        for good in [GOOD_DATA_1, GOOD_DATA_2]:
            good_data = context.parse_and_load(good)
            result = if_true_then_empty(good_data[0].instance, test_schema[0], None, "alpha", "beta")
            self.assertTrue(result.is_success(), f"Expected success for {good_data[0].name}")
            context.remove_definitions(good_data)

    def test_if_true_then_empty_fail(self):

        context = LanguageContext()

        test_schema = context.parse_and_load(TEST_SCHEMA)

        for bad in [BAD_DATA_1]:
            bad_data = context.parse_and_load(bad)
            result = if_true_then_empty(bad_data[0].instance, test_schema[0], None, "alpha", "beta")
            self.assertFalse(result.is_success(), f"Expected failure for {bad_data[0].name}")
            self.assertEqual(result.messages[0].source, "No file to reference")
            context.remove_definitions(bad_data)

        context.remove_definitions(test_schema)


TEST_SCHEMA = """
schema:
    name: IfTrueEmptyTest
    package: test.if_true_then_empty
    root: one
    fields:
        - name: name
          type: string
        - name: alpha
          type: bool
        - name: beta
          type: string[]
    constraints:
        - name: If true then empty
          arguments:
            - name: bool_field_name
              value: alpha
            - name: empty_field_name
              value: beta

"""

GOOD_DATA_1 = """
one:
    name: GoodData1
    alpha: true
"""

GOOD_DATA_2 = """
one:
    name: GoodData2
    alpha: false
    beta:
        - one
        - two
"""

BAD_DATA_1 = """
one:
    name: BadData1
    alpha: true
    beta:
        - one
        - two
"""



"""A simple data class to collect AaC command info from plugins."""

from typing import Any, Callable, Union

from attr import Factory, attrib, attrs, validators
from aac.context.definition import Definition


@attrs
class AacCommandArgument:
    """
    A class used to represent a command argument for an AaC command.

    Attributes:
        name: One or more strings with the name of the command argument.
        description: A string with the command argument description.
        data_type: The data type of the command argument.
        default: The default value for the current argument. (default: None)
    """

    name: Union[str, list[str]] = attrib(validator=validators.instance_of((list, str)))
    description: str = attrib(validator=validators.instance_of(str))
    data_type: str = attrib(validator=validators.instance_of(str))
    default: Any = attrib(default=None)


@attrs(hash=False)
class AacCommand:
    """
    A class used to represent a command in AaC.

    The command name is added to the argument parser with help set to the command description.
    The callback is invoked if the user specifies the command.

    Attributes:
        name: A string with the name of the command.
        description: A string with the command description -- will be used to populate the help command.
        callback: A function that's executed when the user runs the AaC command.
        arguments: A list of AacCommandArgument containing argument information about the command. (default: [])
    """

    name: str = attrib(validator=validators.instance_of(str))
    description: str = attrib(validator=validators.instance_of(str))
    callback: Callable[..., Any] = attrib(validator=validators.is_callable())
    arguments: list[AacCommandArgument] = attrib(
        default=Factory(list), validator=validators.instance_of(list)
    )

    def __hash__(self) -> int:
        """Return the hash of this AacCommand."""
        return hash(self.name)


@attrs(hash=False)
class PluginRunner:
    """
    A class used to manage the plugin execution invocation in AaC.

    Each plugin may define a set of commands and constraints.  The plugin runner contains the plugin definition
    and a mapping of command names to command callbacks and constraint names to constraint callbacks.

    Attributes:
        plugin_definition: The definition for the plugin.
        command_to_callback: Dictionary mapping command names to command callbacks.
        constraint_to_callback: Dictionary mapping constraint names to constraint callbacks.
    """

    plugin_definition: Definition = attrib(validator=validators.instance_of(Definition))
    command_to_callback: dict[str, Callable] = attrib(
        default={}, validator=validators.instance_of(dict)
    )
    constraint_to_callback: dict[str, Callable] = attrib(
        default={}, validator=validators.instance_of(dict)
    )

    # Add constraints here

    def add_command_callback(
        self, command_name: str, command_callback: Callable
    ) -> None:
        """Add a command callback to the plugin runner."""
        if not callable(command_callback):
            raise TypeError("Command Callback is not Callable")
        self.command_to_callback[command_name] = command_callback

    def add_constraint_callback(
        self, constraint_name: str, constraint_callback: Callable
    ) -> None:
        """Add a constraint callback to the plugin runner."""
        if not callable(constraint_callback):
            raise TypeError("Constraint Callback is not Callable")
        self.constraint_to_callback[constraint_name] = constraint_callback

    def get_command_callback(self, command_name: str) -> Callable:
        """Return the callback for the given constraint name."""
        return self.command_to_callback[command_name]

    def get_constraint_callback(self, constraint_name: str) -> Callable:
        """Return the callback for the given constraint name."""
        return self.constraint_to_callback[constraint_name]

    def get_plugin_name(self) -> str:
        """Return the name of the plugin."""
        return self.plugin_definition.name



from unittest import TestCase
from aac.context.source_location import SourceLocation
from aac.context.language_context import LanguageContext


class TestSourceLocation(TestCase):
    def test_source_location(self):
        loc = SourceLocation(1, 2, 3, 4)
        self.assertIsInstance(loc.to_tuple(), tuple)

        loc2 = SourceLocation(2, 1, 3, 4)
        self.assertIsInstance(loc2.to_tuple(), tuple)

        loc3 = SourceLocation(1, 2, 4, 3)
        self.assertIsInstance(loc3.to_tuple(), tuple)
        self.assertNotEqual(loc, loc2, loc3)

    def test_source_location_mapping(self):
        context = LanguageContext()
        definition = context.parse_and_load(VALID_AAC_YAML_CONTENT)[0]
        lexemes = definition.lexemes

        self.assertEqual(lexemes[0].value, "schema")
        self.assertEqual(lexemes[0].location.line, 0)
        self.assertEqual(lexemes[0].location.column, 0)
        self.assertEqual(lexemes[0].location.position, 0)
        self.assertEqual(lexemes[0].location.span, 6)

        self.assertEqual(lexemes[1].value, "name")
        self.assertEqual(lexemes[1].location.line, 1)
        self.assertEqual(lexemes[1].location.column, 2)
        self.assertEqual(lexemes[1].location.position, 10)
        self.assertEqual(lexemes[1].location.span, 4)

    def test_source_location_incorrect_type(self):
        with self.assertRaises(TypeError):
            loc = SourceLocation("1", "2", "3", "4")  # noqa: F841


VALID_AAC_YAML_CONTENT = """
schema:
  name: TestSchema
  description: |
    This is a test schema.
  fields:
    - name: string_field
      type: string
      description: |
        This is a test field.
    - name: integer_field
      type: integer
      description: |
        This is a test field.
    - name: boolean_field
      type: boolean
      description: |
        This is a test field.
    - name: number_field
      type: number
      description: |
        This is a test field.
""".strip()



"""__init__.py module for the Unique Root Keys plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname
from aac.execute.aac_execution_result import (
    ExecutionResult,
)
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.execute.plugin_runner import PluginRunner
from aac.plugins.unique_root_keys.unique_root_keys_impl import plugin_name
from aac.plugins.unique_root_keys.unique_root_keys_impl import root_key_names_are_unique


unique_root_keys_aac_file_name = "unique_root_keys.aac"


def run_root_key_names_are_unique(context: LanguageContext) -> ExecutionResult:
    """Check every definition to ensure there are no duplicate root keys defined in the AaC language."""

    return root_key_names_are_unique(context)


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    unique_root_keys_aac_file = join(dirname(__file__), unique_root_keys_aac_file_name)
    definitions = active_context.parse_and_load(unique_root_keys_aac_file)

    unique_root_keys_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = unique_root_keys_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(plugin_definition=unique_root_keys_plugin_definition)

    plugin_runner.add_constraint_callback(
        "Root key names are unique", run_root_key_names_are_unique
    )

    active_context.register_plugin_runner(plugin_runner)



from unittest import TestCase
from typing import Tuple
from click.testing import CliRunner
from aac.execute.command_line import cli, initialize_cli
from aac.execute.aac_execution_result import ExecutionStatus
from aac.context.language_error import LanguageError

import os
import shutil
import tempfile


from aac.plugins.check.check_aac_impl import plugin_name, check


class TestCheckAaC(TestCase):
    def test_check(self):
        # Let's just test via CLIf for now
        pass

    def run_check_cli_command_with_args(self, args: list[str]) -> Tuple[int, str]:
        """Utility function to invoke the CLI command with the given arguments."""
        initialize_cli()
        runner = CliRunner()
        result = runner.invoke(cli, ["check"] + args)
        exit_code = result.exit_code
        std_out = str(result.stdout)
        output_message = std_out.strip().replace("\x1b[0m", "")
        return exit_code, output_message

    def test_cli_check(self):
        """Test the CLI command for the check plugin."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "good.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            self.assertEqual(0, exit_code, f"Expected success but failed with message: {output_message}")  # asserts the command ran successfully
            self.assertNotIn("My plugin was successful.", output_message) # only appears when --verbose is passed in.

    def test_cli_check_verbose(self):
        """Test the CLI command for the check plugin."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "good.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path, "--verbose"]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            self.assertEqual(0, exit_code, f"Expected success but failed with message: {output_message}")  # asserts the command ran successfully
            self.assertIn("was successful.", output_message) # only appears when --verbose is passed in.

    def test_cli_check_bad_data(self):
        """Test the CLI command for the check plugin."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "bad.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path]
            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            self.assertNotEqual(0, exit_code)
            self.assertIn("was expected to be list, but was", output_message)



"""The AaC Root schema must have name plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

# There may be some unused imports depending on the definition of the plugin...but that's ok
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)
from aac.context.language_error import LanguageError
from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from typing import Any


plugin_name = "Root schema must have name"


def _get_fields(schema) -> list[str]:
    """Returns a list of the fields in the parent schema."""
    context: LanguageContext = LanguageContext()
    fields: list[str] = []
    for field in schema.fields:
        fields.append(field.name)
    if schema.extends:
        for ext in schema.extends:
            parent_schema = context.get_definitions_by_name(ext.name)
            if len(parent_schema) == 1:
                fields.extend(_get_fields(parent_schema[0].instance))
            elif len(parent_schema) > 1:
                raise LanguageError(
                    f"Unable to identify unique definition for {ext.name}:  found {parent_schema} ", "Unknown location"
                )
    return fields


def root_schema_has_name(
    instance: Any, definition: Definition, defining_schema
) -> ExecutionResult:
    """Business logic for the Root schema has name constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    context = LanguageContext()
    if context.is_aac_instance(instance, "aac.lang.Schema"):
        if instance.root:
            # we have one special case for 'import' which is a root without a name, so allow that
            if instance.root == "import":
                return ExecutionResult(
                    plugin_name, "Root schema has name", status, messages
                )

            # get the list of fields in the instance as well as the fields in the parent schema
            fields = _get_fields(instance)  # Could we get the location of the instance and pass it to the function?
            if "name" not in fields:
                status = ExecutionStatus.GENERAL_FAILURE
                error_msg = ExecutionMessage(
                    f"Root schema {instance.name} must have a field named 'name'",
                    MessageLevel.ERROR,
                    "No file to reference" if not definition.source or definition.source.uri == "<string>" else definition.source.uri,
                    None,
                )
                messages.append(error_msg)

    return ExecutionResult(plugin_name, "Root schema has name", status, messages)



"""Python helpers for the generate plugin that can be passed into Jinja2 and used in templates."""
from aac.context.language_context import LanguageContext
from aac import __version__
from typing import Any


# build out jinja2 utility functions
def get_python_name(name: str) -> str:
    """Returns the python name for an AaC name."""
    if name.startswith("--"):
        name = name[2:]
    return name.replace(" ", "_").replace("-", "_").lower()


def get_python_primitive(aac_primitive_name: str) -> str:
    """Returns the python primitive for an AaC primitive."""
    context = LanguageContext()
    aac_primitives = context.get_primitives()
    clean_name = aac_primitive_name
    if clean_name.endswith("[]"):
        clean_name = clean_name[:-2]
    if clean_name.find("(") > -1:
        clean_name = clean_name[: clean_name.find("(")]
    for aac_primitive in aac_primitives:
        if clean_name == aac_primitive.name:
            return aac_primitive.instance.python_type
    return "None"


def get_python_type(aac_type_name: str) -> str:
    """Returns the python type for an AaC type."""
    context = LanguageContext()
    if aac_type_name.endswith("[]"):
        aac_type_name = aac_type_name[:-2]
        return f"list[{get_python_type(aac_type_name)}]"
    elif aac_type_name in [primitive.name for primitive in context.get_primitives()]:
        return context.get_python_type_from_primitive(aac_type_name)
    else:
        return aac_type_name


def get_package_from_plugin(plugin_name: str) -> str:
    """Returns the package for an AaC plugin."""
    context = LanguageContext()
    plugin_runners = context.get_plugin_runners()
    for runner in plugin_runners:
        if plugin_name == runner.plugin_definition.name:
            return runner.plugin_definition.instance.package
    return f"aac.plugins.{get_python_name(plugin_name)}"


def get_package_from_aac_definition(item_name: str) -> str:
    """Returns the package for an AaC definition."""
    context = LanguageContext()
    definitions = context.get_definitions()
    for definition in definitions:
        if definition.name == item_name:
            return definition.instance.package
    return ""


def get_path_from_package(package: str) -> str:
    """Converts a package to a path."""
    return package.replace(".", "/")


def aac_version() -> str:
    """Returns the current version of AaC."""
    return __version__


def _generate_test_data_for_primitive(primitive_type: str) -> Any:
    if primitive_type == "str":
        return "test"
    elif primitive_type == "int":
        return 1
    elif primitive_type == "float":
        return 1.1
    elif primitive_type == "bool":
        return True
    elif primitive_type == "Any":
        return "{}"
    else:
        raise Exception(f"Unknown primitive type: {primitive_type}")


def schema_to_test_dict(name: str, omit_optional_fields: bool = False) -> dict:
    """Generate test data for a schema to a dictionary based on its fields."""
    context: LanguageContext = LanguageContext()
    schema_definitions = context.get_definitions_by_name(name)
    if len(schema_definitions) != 1:
        raise Exception(f"Could not find unique schema definition: {name}")
    schema_definition = schema_definitions[0]
    result: dict = {}

    # we need to avoid the case where a schema is defining an AaC type other than Schema
    if schema_definition.get_root_key() == "enum":
        # need to return a valid enum value
        for value in schema_definition.instance.values:
            return value

    for field in schema_definition.instance.fields:
        if omit_optional_fields and not field.is_required:
            continue

        clean_field_type = get_python_primitive(field.type)
        if clean_field_type == "None":
            # this is a schema type
            if field.type.endswith("[]"):
                # this is an array of schema types
                result[field.name] = [
                    schema_to_test_dict(field.type[:-2], omit_optional_fields),
                    schema_to_test_dict(field.type[:-2], omit_optional_fields),
                ]
            else:
                result[field.name] = schema_to_test_dict(
                    field.type, omit_optional_fields
                )
        else:
            # this is a primitive type
            if field.type.endswith("[]"):
                # this is an array of primitive types
                result[field.name] = [
                    _generate_test_data_for_primitive(clean_field_type),
                    _generate_test_data_for_primitive(clean_field_type),
                ]
            else:
                result[field.name] = _generate_test_data_for_primitive(clean_field_type)

    return result
